package org.eso.dfs.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectStreamClass;

/**
 * @author nkornwei
 *
 * This class was created as part of refactoring the class structure, 
 * in order to maintain backwards compatibility with serialized
 * objects on user file systems.
 *  
 * This ObjectInput stream maps a certain package class path to another
 * package class path.
 * 
 * All serialized objects in the org.eso.gasgano package are now a part
 * of org.eso.gasgano.keyword.
 * This class checks that Class definitions are mapped when appropriate
 * when an object stream is being read.
 * 
 */

public class UpgradingObjectInputStream extends ObjectInputStream {

    /**
     * The original package path of serialized objects.
     */
    private String oldPackage;

    /**
     * The new package path of serialized objects (the package containing
     * the classes serialized objects are to be mapped to).
     */
    private String newPackage;

    /**
     * Constructor for UpgraderObjectInputStream.
     * @param in
     * @param oldpkg the packahe path expected from the stream. E.g. "org.eso.gasgano.". Note
     * the "." at the end.
     * @param newpkg the new package path to map to, E.g. "org.eso.gasgano.keyword.". Note
     * the "." at the end.
     * @throws IOException
     */
    public UpgradingObjectInputStream(
        InputStream in,
        String oldpkg,
        String newpkg)
        throws IOException {
        super(in);
        oldPackage = oldpkg;
        newPackage = newpkg;
    }

    /**
     * Constructor for UpgraderObjectInputStream.
     * @throws IOException
     * @throws SecurityException
     */
    public UpgradingObjectInputStream() throws IOException, SecurityException {
        super();
    }

    /**
     * Read in a Class Descriptor from the stream - loading a different
     * class descriptor if the class has a specific package name.
     * Does nothing if the old or new package names are null.
     * @see java.io.ObjectInputStream#readClassDescriptor()
     */
    protected ObjectStreamClass readClassDescriptor()
        throws IOException, ClassNotFoundException {
        ObjectStreamClass osc = super.readClassDescriptor();
        if (oldPackage != null && newPackage != null) {
            if (osc.getName().startsWith(oldPackage)
                && !osc.getName().startsWith(newPackage)) {
                String className = osc.getName().substring(oldPackage.length());
                System.out.println(
                    "Upgrading: "
                        + osc.getName()
                        + " to "
                        + newPackage
                        + className);
                osc =
                    ObjectStreamClass.lookup(
                        Class.forName(newPackage + className));
            }
        }
        return osc;
    }
    /**
     * Returns the newPackage.
     * @return String
     */
    public String getNewPackage() {
        return newPackage;
    }

    /**
     * Returns the oldPackage.
     * @return String
     */
    public String getOldPackage() {
        return oldPackage;
    }

    /**
     * Sets the newPackage.
     * @param newPackage The newPackage to set
     */
    public void setNewPackage(String newPackage) {
        this.newPackage = newPackage;
    }

    /**
     * Sets the oldPackage.
     * @param oldPackage The oldPackage to set
     */
    public void setOldPackage(String oldPackage) {
        this.oldPackage = oldPackage;
    }

}
