/**
 * These classes have been developed for the DFSLog system.
 *
 * @author N. Kornweibel nkornwei@eso.org
 *
 * TO DO:
 * need @revision, and SNIFF info here.
 * 
 */

package org.eso.dfslog;

import java.io.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;

/**
 * ReportViewer:
 *
 * A simple Frame for viewing, editing, saving and printing the output
 * from objects implementing the Report interface.
 *
 * The class uses a Report interface to query report related
 * information (only column widths at the moment).
 */


public class ReportViewer extends JFrame implements ActionListener {

    private String fileName;
    private JTextArea ta;
    private Font printerFont = null;

    private int columnWidths[] = null;


    private void setColumnWidths(int c[]) {
        columnWidths = new int[c.length];
        int total = 0;
        for (int i = 0; i < c.length; i++) {
            total = total+c[i]+2;
            columnWidths[i] = total;
        }
    }



    /**
     * Constructor creates a temporary file with a PrintWriter,
     * generates the report and reads the report from the
     * temporary file.
     */

    public ReportViewer(String title, Report rep) {
        super(title);
        Random rnd = new Random();
        fileName = new String("/tmp/dfsreport"+rnd.nextInt());
        String errorString = null;
        boolean reportCancelled = false;
        try {
            FileWriter fw = new FileWriter(fileName);
            PrintWriter pw = new PrintWriter(fw, true);
            rep.setPrintWriter(pw);

            ReportRunner runner = new ReportRunner(rep);

            ReportDialog runReport = new ReportDialog(this,
                "Report", "Generating Report...", true, runner);
            runReport.run();
            reportCancelled = runReport.cancelSelected();
            if (reportCancelled) {
                    errorString = new String("Report cancelled.");
            } else {
                if (runner.reportStatus() == false) {
                    errorString = new String("Error generating report.");
                } else {
                    setColumnWidths(rep.getColumnWidths());
                }
                pw.flush();
                pw.close();
            }
        } catch (IOException e) {
            errorString = new String("Error generating report: "+e.toString());
        }
        if (errorString != null) {
            JOptionPane.showMessageDialog(this, errorString);
        }

        // open the file and read in the contents.
        // this section will produce memory problems when attempting to 
        // show very large reports.
        // Although it is unusual to catch an OutOfMemory error, we will
        // in this case to display an error message and correctly clean
        // up the report.
  
        BufferedReader in;
        try {
            in = new BufferedReader(new FileReader(fileName));
            StringBuffer text = new StringBuffer();
            String s = "";
            while ((s = in.readLine()) != null) {
                text.append("\n"+s);
            }
            ta = new JTextArea(text.toString());
            // set small Font size.
            Font taFont = new Font("Courier", Font.PLAIN, 10);
            ta.setFont(taFont);
         
            ta.setEditable(false);
            getContentPane().add(new JScrollPane(ta));
        } catch (OutOfMemoryError e) {
            JOptionPane.showMessageDialog(this, 
                "Error reading file: "+fileName
                +"\n the file is too large for display.");
        } catch (IOException e) {
            JOptionPane.showMessageDialog(this, 
                "Error reading file: "+fileName);
        }
    
        // Setup the menu items....
        MenuBar mb = new MenuBar();
        Menu fileMenu = new Menu("File");
        MenuItem printMenuItem = new MenuItem("Print");
        printMenuItem.addActionListener(this);
        fileMenu.add(printMenuItem);
        MenuItem saveAsMenuItem = new MenuItem("Save As...");
        saveAsMenuItem.addActionListener(this);
        fileMenu.add(saveAsMenuItem);
        MenuItem quitMenuItem = new MenuItem("Quit");
        quitMenuItem.addActionListener(this);
        fileMenu.add(quitMenuItem);
        mb.add(fileMenu);

        Menu fontMenu = new Menu("Display Size");
        MenuItem fontItem = new MenuItem("Small");
        fontItem.addActionListener(this);
        fontMenu.add(fontItem);
        fontItem = new MenuItem("Large");
        fontItem.addActionListener(this);
        fontMenu.add(fontItem);
        mb.add(fontMenu);

        setMenuBar(mb);
        printerFont = new Font("Courier", Font.PLAIN, 3);
    }




    // implementation of the ActionListener interface
    public void actionPerformed(ActionEvent a) {
        if (a.getActionCommand() == "Quit") {
            File f = new File(fileName);
            boolean deleted = true;
            try {
                deleted = f.delete();
            } catch (SecurityException e) {
                deleted = false;
            }
            if (!deleted) {
                JOptionPane.showMessageDialog(this,
                    "Failed to delete temporary report file:\n"+fileName);
            }
            setVisible(false);
            dispose();
        } else if (a.getActionCommand() == "Save As...") {
/*    Old FileDialog Code:
            FileDialog fd = new FileDialog(this, "Save report to file",
                FileDialog.SAVE);
            fd.show();

            if (fd.getFile() != null && 
                saveFile(fd.getDirectory()+fd.getFile())) {
                JOptionPane.showMessageDialog(this,
                    "Report saved to: "+fd.getDirectory()+fd.getFile());
            }
*/
            JFileChooser fc = new JFileChooser();
            fc.setDialogTitle("Enter tar file name");
            fc.setDialogType(JFileChooser.SAVE_DIALOG);
            String selectedFile = null;
            if (fc.showSaveDialog(this) == JFileChooser.APPROVE_OPTION) {
                selectedFile = fc.getSelectedFile().getPath();
            }
            if (selectedFile != null && saveFile(selectedFile)) {
                JOptionPane.showMessageDialog(this,
                    "Report saved to: "+selectedFile);
            }
        } else if (a.getActionCommand() == "Print") {
            Properties p = new Properties();
            // the following properties apply only to UNIX java VM.
            // need to put these in general properties box.
            p.put("awt.print.numCopies", "1");
            p.put("awt.print.paperSize", "A4");
            p.put("awt.print.destination", "printer");
            p.put("awt.print.orientation", "landscape");
            p.put("awt.print.printer", "lp");
            PrintJob pjob = getToolkit().getPrintJob(this, "Test", p);
            if (pjob != null) {
                Graphics pg = pjob.getGraphics();
                if (pg != null) {
                    pg.setFont(printerFont);
                    int lineHeight = pg.getFontMetrics().getAscent()
                        +pg.getFontMetrics().getDescent();
                    BufferedReader in;
                    int charWidth = pg.getFontMetrics().charWidth('X');
                    try {
                        in = new BufferedReader(new FileReader(fileName));
                        String s = "";
                        int lines = 2;
                        int pageNum = 1;
                        int height = pjob.getPageDimension().height;
                        while ((s = in.readLine()) != null) {
                            lines++;
                            if ((lines+2)*lineHeight > height) {
                                pageNum++;
                                pg.dispose();
                                pg = pjob.getGraphics();
                                pg.setFont(new Font("Courier", Font.PLAIN, 3));
                                lines = 4;
                                printString("Page: "
                                    +pageNum, pg, 2*lineHeight, charWidth);
                            }
                            printString(s, pg, lines*lineHeight, charWidth);
                        }
                    } catch (IOException e) {
                        JOptionPane.showMessageDialog(this,
                            "Error reading file: "+fileName);
                    }
                    pg.dispose();
                }
                pjob.end();
            }
        } else if (a.getActionCommand() == "Small") {
            Font taFont = new Font("Courier", Font.PLAIN, 7);
            ta.setFont(taFont);
        } else if (a.getActionCommand() == "Large") {
            Font taFont = new Font("Courier", Font.PLAIN, 10);
            ta.setFont(taFont);
        }
    }




    private void printString(String s, Graphics pg, int height, int width) {
        int fromIndex = 0;
        int toIndex = 0;
        int colIndex = -1;
        do {
            toIndex = s.indexOf('\t', fromIndex);
            String field;
            try {
                if (toIndex == -1) {
                    field = s.substring(fromIndex);
                } else {
                    field = s.substring(fromIndex, toIndex);
                }
            } catch (StringIndexOutOfBoundsException e) {
                    field = null;
            }
            if (field != null) {
                int offset = 30;
                if (colIndex >= 0 && columnWidths != null
                    && colIndex < columnWidths.length) {
                        offset = offset+(columnWidths[colIndex]*width);
                }
                pg.drawString(field, offset, height);
            }
            fromIndex = toIndex+1;
            colIndex++;
        } while (toIndex != -1);
    }




    private int[] setupColumns(String s, FontMetrics fm) {
        int columns[] = null;
        int nCols = 1;
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == '\t') {
                nCols++;
            }
        }
        columns = new int[nCols];
        int from = 0;
        int colIndex = 0;
        int total = 0;
        for (int i = 0; i < s.length(); i++) {
            if (s.charAt(i) == '\t') {
                int columnWidth = fm.stringWidth("X")*((i - from));
                columns[colIndex] = total+columnWidth;
                total = total+columnWidth;
                colIndex++;
                from = i;
            }
        }
        return columns;
    }




    boolean saveFile(String name) {
        boolean rc = true;
        if (name == null) {
            return false;
        }
        try {
            FileWriter fw = new FileWriter(name);
            PrintWriter pw = new PrintWriter(fw);
            FileReader fr = new FileReader(fileName);
            BufferedReader br = new BufferedReader(fr);
            String s = br.readLine();
            while (s != null) {
                pw.println(s);
                s = br.readLine();
            }
            pw.flush();
            pw.close();
        } catch (IOException e) {
            rc = false;
            JOptionPane.showMessageDialog(this,
                "Error saving report: "+e.toString());
        }
        return rc;
    }
} 
// end of file ReportViewer.java.
