package org.eso.gasgano;

/**
 * @author Nick Kornweibel
 * 
 * This class presents a window of scripts names in a table. The scripts (table cells) 
 * may be selected through a double click, and executed.
 * 
 * The window, is visible, is sensitive to changes in the file display tree of
 * Gasgano.
 * 
 * This class is in org.eso.gasgano because it has a dependecy on GasganoDisplay.
 * TODO: remove the dependecy on GasganoDisplay.
 */

import jal.String.Sorting;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FilenameFilter;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.border.EmptyBorder;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.table.AbstractTableModel;

import org.eso.gasgano.gui.DFSFileTreeTable;

public class ScriptDirBoard
    extends JFrame
    implements ActionListener, TreeSelectionListener, ComponentListener {

    private boolean isBeingShown = false;
    private JFileChooser fc = null;
    private ScriptBoardModel scriptTableModel = null;
    private JTable scriptTable = null;
    private JPanel scriptPanel = null;
    private JPanel buttonPanel = null;
    private JPanel radioPanel = null;
    private JButton dismissButton = null;
    private JButton overrideButton = null;
    private JRadioButton confirmButton = null;
    private JRadioButton classificationSupplyButton = null;
    private JRadioButton backGroundButton = null;

    // settings for controlling command line parameters and command execution
    private boolean confirmExec = true;
    private boolean classificationSupply = false;
    private boolean backGround = false;

    private JScrollPane scrollPane = null;
    private String[] fileList = null;
    private String selectedDir = null;
    private String baseDir = null;
    private String instrumentName = null;

    private File scriptDir = null;
    private Dimension recordedBounds = null;
    private GasganoDisplay gasganoDisplay = null;
    private DFSFileTreeTable fsTree = null;

    private boolean showingDefault = false;

    /**
     * Constructor
     * 
     * @param baseDir
     * @param gasganoDisplay
     * @param fsTree
     */
    public ScriptDirBoard(
        String baseDir,
        GasganoDisplay gasganoDisplay,
        DFSFileTreeTable fsTree) {

        super("Gasgano Script Directory Board");
        this.baseDir = baseDir;
        this.gasganoDisplay = gasganoDisplay;
        this.fsTree = fsTree;
        selectedDir = baseDir;
        if (selectedDir == null) {
            System.out.println("User selected null file?");
            return;
        }
        scriptDir = new File(selectedDir);
        instrumentName = "";
        // we should get only plain files though
        fileList = scriptDir.list(new NoDirsFilter());

        // create the table
        scriptTableModel = new ScriptBoardModel(fileList);
        scriptTable = new JTable(scriptTableModel);

        scriptTable.setCellSelectionEnabled(true);
        scriptTable.setSelectionMode(
            ListSelectionModel.SINGLE_INTERVAL_SELECTION);
        scriptTable.setAutoResizeMode(JTable.AUTO_RESIZE_ALL_COLUMNS);
        scriptTable.setPreferredScrollableViewportSize(new Dimension(240, 100));

        scriptTable.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent evt) {
                if (evt.getClickCount() == 2) {
                    Point pt = evt.getPoint();
                    int rowIndex = scriptTable.rowAtPoint(pt);
                    int columnIndex = scriptTable.columnAtPoint(pt);
                    if (rowIndex == -1 || columnIndex == -1) {
                        System.out.println("Clicked outside table");
                    } else {
                        String commandString =
                            new String(
                                selectedDir
                                    + File.separator
                                    + (String) scriptTable.getModel().getValueAt(
                                        rowIndex,
                                        columnIndex));
                        String completeCommandPath =
                            new File(commandString).getAbsolutePath();
                        ScriptDirBoard
                            .this
                            .gasganoDisplay
                            .selectedFilesToCommand(
                            completeCommandPath,
                            confirmExec,
                            classificationSupply,
                            backGround);
                    }
                }
            }
        });

        dismissButton = new JButton("Dismiss");
        dismissButton.setSize(dismissButton.getMinimumSize());
        dismissButton.addActionListener(this);

        // NB: the name is configured in configureScriptButton()
        overrideButton = new JButton("use default scripts");
        overrideButton.setSize(overrideButton.getMinimumSize());
        overrideButton.addActionListener(this);

        confirmButton = new JRadioButton("confirm");
        confirmButton.setSelected(confirmExec);

        confirmButton.addActionListener(this);
        classificationSupplyButton = new JRadioButton("Supply classification");
        classificationSupplyButton.addActionListener(this);
        classificationSupplyButton.setSelected(classificationSupply);

        backGroundButton = new JRadioButton("Run in background");
        backGroundButton.addActionListener(this);
        backGroundButton.setEnabled(false);
        backGroundButton.setSelected(backGround);

        // create the panels
        scriptPanel = new JPanel(new BorderLayout());
        buttonPanel = new JPanel(new BorderLayout());
        radioPanel = new JPanel(new GridBagLayout());
        // create scrollPane
        scrollPane = new JScrollPane(scriptTable);

        // add scrollPane to Panel
        scriptPanel.add(scrollPane, BorderLayout.CENTER);
        //scriptPanel.add(scriptTable,BorderLayout.CENTER);
        scriptPanel.setBorder(
            new TitledBorder(new EtchedBorder(), "Double-click to execute"));

        radioPanel.add(confirmButton);
        radioPanel.add(classificationSupplyButton);
        radioPanel.add(backGroundButton);
        radioPanel.setBorder(new EmptyBorder(2, 5, 2, 5));

        buttonPanel.add(dismissButton, "East");
        buttonPanel.add(overrideButton, "West");
        buttonPanel.setBorder(new EmptyBorder(5, 5, 5, 5));

        getContentPane().setLayout(new BorderLayout());
        getContentPane().add(scriptPanel, BorderLayout.CENTER);
        getContentPane().add(buttonPanel, BorderLayout.NORTH);
        getContentPane().add(radioPanel, BorderLayout.SOUTH);

        //   setResizable(false);

        setDefaultCloseOperation(JFrame.DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent evt) {
                setVisible(false);
            }
        });
        pack();
        addComponentListener(this);
        // update the model and display to show the current selection.
        // this should be better integrated into the code above - so the
        // model and display is not built twice.
        displayScripts();
    }

    // method for the action listener 
    public void actionPerformed(ActionEvent ae) {
        if (ae.getSource() == overrideButton) {
            showingDefault = !showingDefault;
            displayScripts();
        }

        if (ae.getSource() == dismissButton) {
            setVisible(false);
        }
        if (ae.getSource() == confirmButton) {
            confirmExec = confirmButton.isSelected();
        }

        if (ae.getSource() == classificationSupplyButton) {
            classificationSupply = classificationSupplyButton.isSelected();
        }
        if (ae.getSource() == backGroundButton) {
            backGround = backGroundButton.isSelected();
        }
    }

    /**
     * 
     * @return String
     */
    private void configureScriptButton() {
        SwingUtilities.invokeLater(new Runnable() {
            public void run() {
                overrideButton.setEnabled(instrumentName != null);
                String name = "";
                if (instrumentName != null) {
                    name =
                        showingDefault
                            ? "Show " + instrumentName + " Scripts"
                            : "Show Default Scripts";
                }
                overrideButton.setText(name);
            }
        });

    }
    /**
     * method for the TreeSelectionListener()<p>
     * here we watch to see whether the selected instrument has changed
     * and take care of the possibility that the corresponding directory
     * may not exist<p>
     */

    public void valueChanged(TreeSelectionEvent e) {
        if (fsTree.getSelectedOBComponents().size() == 0) {
            return;
        }
        displayScripts();
    }

    /**
     * Check the fsTree current selection and display the appropriate 
     * scripts (based on the instrument and the current showingDefault setting).
     */
    public void displayScripts() {
        instrumentName = fsTree.getSelectedInstrument();
        if (!canDisplayInstrumentScripts()) {
            instrumentName = null;
        }
        if (showingDefault || instrumentName == null) {
            displayDefaultScripts();
        } else {
            displayInstrumentScripts();
        }
        configureScriptButton();
    }

    /**
     * Determine if the current instrumentName setting is a
     * valid script directory.
     * @return boolean
     */
    private boolean canDisplayInstrumentScripts() {
        if (instrumentName == null) {
            return false;
        }
        String dir = baseDir + File.separator + instrumentName;
        File scrDir = new File(dir);
        return scrDir.exists() && scrDir.isDirectory();
    }

    /**
     * resets some status variables and sets the border
     * to a string without instrument name<p>
     * this method assumed canDisplayInstrumentScripts() returns true.
     */
    private void displayInstrumentScripts() {
        selectedDir = baseDir + File.separator + instrumentName;
        scriptDir = new File(selectedDir);
        scriptPanel.setBorder(
            new TitledBorder(
                new EtchedBorder(),
                "Double-click to execute (" + instrumentName + ")"));

        if (isVisible()) {
            showBoard();
            //pack();
            scriptTableModel.fireTableStructureChanged();
            scriptTable.sizeColumnsToFit(-1);
        }
        showingDefault = false;
    }

    /**
     * resets some status variables and sets the border
     * to a string without instrument name<p>
     */
    private void displayDefaultScripts() {
        selectedDir = baseDir;
        scriptDir = new File(selectedDir);
        scriptPanel.setBorder(
            new TitledBorder(new EtchedBorder(), "Double-click to execute "));
        if (isVisible()) {
            showBoard();
            scriptTableModel.fireTableStructureChanged();
            scriptTable.sizeColumnsToFit(-1);
        }
        showingDefault = true;
    }

    public void showBoard() {
        setVisible(true);
        scriptTableModel.loadFiles(scriptDir.list(new NoDirsFilter()));
        if (recordedBounds != null) {
            setSize(recordedBounds);
        }
        //scriptTable.repaint();
    }

    public void hideBoard() {
        // scriptTableModel.loadFiles(scriptDir.list(new NoDirsFilter() ) ); 
        setVisible(false);
    }

    public void componentResized(ComponentEvent e) {
        recordedBounds = getSize();
    }

    public void componentMoved(ComponentEvent e) {
    }

    public void componentShown(ComponentEvent e) {
    }

    public void componentHidden(ComponentEvent e) {
    }

    /**
     * Inner class for selecting only files (not directories)
     */
    private static class NoDirsFilter implements FilenameFilter {
        public boolean accept(File dir, String name) {
            if (name.startsWith(".")) {
                return false;
            }
            if (new File(dir.toString() + File.separator + name)
                .isDirectory()) {
                return false;
            }
            return true;
        }
    }

    /**
     * Table Model for displaying an array of scripts. The model accepts
     * an array, and maps the array into a "block" of cells, for best displaying
     * the set of scripts.
     */
    private static class ScriptBoardModel extends AbstractTableModel {
        /**
         * The maximum number of scripts to place in a single column, when reached
         * the next column is used (i.e a column is added to the table model).
         */
        private final int maxOneCol = 10;
        private int columns = 0;
        private int rows = 0;
        private int old_columns, old_rows = 0;

        private Vector fileListLabel = new Vector();

        // constructor 
        public ScriptBoardModel(String[] fileList) {
            loadFiles(fileList);
        }

        public int getColumnCount() {
            return columns;
        }
        public int getRowCount() {
            return rows;
        }

        public Object getValueAt(int row, int col) {
            return fileListLabel.elementAt(row * columns + col);
        }

        public boolean isCellEditable(int row, int col) {
            return false;
        }

        public String getColumnName(int column) {
            return new String("");
        }

        public void setValueAt(Object aValue, int row, int column) {
        }
        public void removeRow(int rowNum) {
        }
        public void insertRow(int rowNum) {
        }

        public void loadFiles(String[] fileList) {
            Sorting.sort(fileList, 0, fileList.length);

            if (fileList.length < maxOneCol) {
                columns = 1;
                rows = fileList.length;
            } else {
                // want to make a "block" of table cells, be spearding the
                // array over columns and rows.
                columns =
                    Math.min((int) Math.sqrt((double) fileList.length), 6);
                rows = fileList.length / columns;
                if (rows * columns < fileList.length) {
                    rows++;
                }
            }
            fileListLabel.removeAllElements();
            fileListLabel.setSize(rows * columns);
            for (int i = 0; i < fileList.length; i++) {
                fileListLabel.setElementAt(fileList[i], i);
            }
            for (int i = fileList.length; i < rows * columns; i++) {
                fileListLabel.setElementAt("", i);
            }
            //System.out.println("Rows "+rows+" Columns "+columns+"  ("+old_rows+" " + old_columns+" ) ") ;
            fireTableDataChanged();
            if (old_columns != columns || old_rows != rows) {
                fireTableStructureChanged();
            }
            old_columns = columns;
            old_rows = rows;
        }
    }

}
