/**
 * @author N. Kornweibel
 * @version 1.0
 * The DataModel represents a set of files structured into programs,
 * and observations. The DataModel is the primary API into the set of
 * files, maintaining indexes, adding and removing files and updating the set
 * to reflect new data.
 * 
 * The DFSDataModel implements the AssociationHandler interface allowing
 * it to act as an AssociationService in Gasgano. By default it
 * is NOT registered as an AssociationService this must be done by the calling
 * application.
 * 
 */

package org.eso.gasgano.datamodel.filesystem;

import java.io.File;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.eso.dfs.services.ServiceException;
import org.eso.dfs.services.data.FileAssociationNode;
import org.eso.dfs.services.data.AssociationNode;
import org.eso.dfs.services.data.AssociationHandler;
import org.eso.gasgano.datamodel.DataModelChangeNotifier;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.FileDescription;
import org.eso.gasgano.datamodel.FrameInfo;
import org.eso.gasgano.datamodel.InterContainer;
import org.eso.gasgano.datamodel.OBComponentFilter;
import org.eso.gasgano.datamodel.Observation;
import org.eso.gasgano.datamodel.ObservationDescription;
import org.eso.gasgano.datamodel.Program;
import org.eso.gasgano.datamodel.ProgramDescription;
import org.eso.gasgano.datamodel.ReductionBlockCommon;
import org.eso.gasgano.datamodel.gui.DFSTextFileOBComponent;
import org.eso.gasgano.datamodel.gui.FitsFileOBComponent;
import org.eso.gasgano.datamodel.gui.KeywordContainerOBComponent;
import org.eso.gasgano.datamodel.gui.OBComponent;
import org.eso.gasgano.datamodel.gui.ReductionBlockOBComponent;
import org.eso.gasgano.datamodel.gui.TFitsFileOBComponent;
import org.eso.gasgano.datamodel.gui.VirtualReductionBlockOBComponent;
import org.eso.gasgano.keyword.RuleTable;
import org.eso.gasgano.properties.GasProp;
import org.eso.gasgano.properties.PropertyDB;
import org.eso.gasgano.tools.SortedHashtable;

public class DFSDataModel
    extends DataModelChangeNotifier
    implements AssociationHandler {

    /**
     * Vector of FileSelectionModel objects that, together, form
     * the full set of files in the data model.
     */
    private Vector fileSelectionModels;
    
    /**
     * The file selection model used to maintain files added through
     * the Gasgano Data Service - i.e. added through the AssociationHandler.
     */
    private ServiceFileSet serviceFileModel;

    /**
     * A filter that every OBComponent is passed through before
     * determining if it is a part of the model or not. If null all
     * OBComponents (files) are included in the model.
     */
    private OBComponentFilter fileFilter;

    /**
     * Code level setting to fix behaviour of the datamodel.
     * Indicates if the ARC_NAME_INDEX index is to be used or not.
     */
    private static final boolean useArcNameIndex = true;

    /**
     * set of programs, each program contains obs, which contains components.
     */
    private SortedHashtable programs;

    /**
     * set of directories, each contains programs,
     * that contain obs that contain components.
     */
    private SortedHashtable directories;

    /**
     * set or telescopes that contain programs
     * that contain obs that containing components.
     */
    private SortedHashtable telescopes;

    /**
     * Factory class for generating descriptions for Programs.
     */
    private ProgramDescription progDescFactory;

    public static final String UNKNOWN_PROGRAM_ID = "Unknown Program";
    public static final String UNKNOWN_OBSERVATION_ID = "Unknown Observation";
    public static final String UNKNOWN_TELESCOPE_ID = "Unknown Telescope";

    /**
     * Integers used by the DataModelBuilder class to display the current
     * status of the file update() process (when running in a 
     * seperate thread).
     */
    private int totalFilesInSet = 0;
    private int numberOfFilesUpdated = 0;

    /**
     * The set of Elementindexes maintained by this DataModel.
     * This array will be of size NUMBER_OF_INDEXES.
     */
    private Hashtable indexes[];

    /**
     * The number of indexes maintained by this DataModel.
     */
    private static final int NUMBER_OF_INDEXES = 6;

    /**
     * 
     * The constants defining index locations in the indexes array.
     * 
     * NAME_INDEX
     * An index of OBComponents based on the name
     * 
     * BASE_NAME_INDEX
     * An index of OBComponents based on the base name - as 
     * a base name is not unique, each element in the index is an
     * array (Vector) of OBComponents.
     * 
     * ARC_NAME_INDEX
     * An index of OBComponents based on the arc name
     * 
     * FS_BASE_NAME_INDEX
     * An index of OBComponents based on the filesystem basename
     * 
     * MD5_INDEX
     * An index of OBComponents based on their MD5 signature
     * 
     * VRB_INDEX
     * An index of VirtualReductionBlockOBComponents
     * based on their product basename. This index is separate from the
     * other 5 indexes - it is for VirtualReductionBlock objects only.
     * 
     */
    private static final int NAME_INDEX = 0;
    private static final int BASE_NAME_INDEX = 1;
    private static final int ARC_NAME_INDEX = 2;
    private static final int FS_BASE_NAME_INDEX = 3;
    private static final int MD5_INDEX = 4;
    private static final int VRB_INDEX = 5;

    /**
     * Classification rules used by the data model.
     */
    private RuleTable ruleTable;

    /**
     * Singleton Pattern instance.
     * In this implementation there can only be one DataModel object 
     * - this is it.
     */
    private static DFSDataModel theDataModel;

    /**
     * Setting for controlling event transmission when removing files.
     * If large numbers of files are removed, even transmission is suspended
     * until the operation is complete.
     */
    private final int eventTolerance = 30;

    /**
     * Is the datamodel being updated/rebuilt currently.
     */
    private boolean dataModelUpdating = false;

    /**
     * Constructor to create a data model. This is private as part of
     * the Singleton Pattern.
     * The DFSDataModel instance is obtained through DFSDataModel.getDataModel().
     */
    private DFSDataModel() {
        fileSelectionModels = new Vector();
        serviceFileModel = new ServiceFileSet();
        fileSelectionModels.add(serviceFileModel);
        
        programs = new SortedHashtable();
        directories = new SortedHashtable();
        telescopes = new SortedHashtable();
        indexes = new Hashtable[NUMBER_OF_INDEXES];
        for (int i = 0; i < NUMBER_OF_INDEXES; i++) {
            indexes[i] = new Hashtable();
        }
    }

    /**
     * Utility method to print structure sizes.
     */
    public void displayHashtableSizes() {
        System.out.println("");
        System.out.println(
            " Hashtables for names : "
                + indexes[NAME_INDEX].size()
                + " "
                + indexes[BASE_NAME_INDEX].size()
                + " "
                + indexes[ARC_NAME_INDEX].size()
                + " "
                + indexes[FS_BASE_NAME_INDEX].size());
        System.out.println(
            " Telescopes: "
                + telescopes.size()
                + " Directories "
                + directories.size()
                + " programs "
                + programs.size());
        System.out.println("");
    }

    /**
     * @param p The ProgramDescription object to be used to configure the
     * descriptions of Program managed by this DataModel.
     */
    public void setDefaultProgramDescription(ProgramDescription p) {
        progDescFactory = p;
    }

    /**
     * @return The ProgramDescription object to be used to configure the
     * descriptions of Program managed by this DataModel.
     */
    public ProgramDescription getDefaultProgramDescription() {
        return progDescFactory;
    }

    /**
     * @param p The ObservationDescription object to be used to configure the
     * descriptions of observation objects managed by this DataModel.
     */
    public void setDefaultObservationDescription(ObservationDescription p) {
        Program.setDefaultObservationDescription(p);
    }

    /**
     * @return The ObservationDescription object to be used to configure the
     * descriptions of observation objects managed by this DataModel.
     */
    public ObservationDescription getDefaultObservationDescription() {
        return Program.getDefaultObservationDescription();
    }

    /**
     * @param p The FileDescription object to be used to configure the
     * descriptions of OBComponent objects managed by this DataModel.
     */
    public void setDefaultFileDescription(FileDescription fd) {
        OBComponent.setDefaultFileDescription(fd);
    }

    /**
     * @return The FileDescription object to be used to configure the
     * descriptions of OBComponent objects managed by this DataModel.
     */
    public FileDescription getDefaultFileDescription() {
        return OBComponent.getDefaultFileDescription();
    }

    /**
     * Get the singleton instance of the DFSDataModel.
     * @return The global DataModel object.
     */
    public static synchronized DFSDataModel getDataModel() {
        if (theDataModel == null) {
            theDataModel = new DFSDataModel();
        }
        return theDataModel;
    }

    /**
     * @param m The FileSelectionModel to be added to this DataModel.
     */
    public void addFileSelectionModel(FileSelectionModel m) {
        fileSelectionModels.add(m);
    }

    /**
     * @param m The FileSelectionModel to be removed from this DataModel.
     */
    public void removeFileSelectionModel(FileSelectionModel m) {
        fileSelectionModels.remove(m);
    }

    /**
     * @param name The id name of the OBComponent to be located.
     * @return The single OBComponent located (or null for a failed search).
     * This method uses the basename only part of the file.
     */
    public OBComponent getObComponentByFilesystemBaseName(String name) {
        OBComponent rc = null;
        if (name != null) {
            rc = (OBComponent) indexes[FS_BASE_NAME_INDEX].get(name);
        }
        return rc;
    }

    /**
     * @param name The base name of the OBComponents to be located.
     * @return Finding an OBComponent by its base name will return a 
     * Vector of components (as multiple components can share
     * the same base name) or null if none are located.
     * Note: an empty Vector will never be returned, but elements
     * in the returned Vector may be null.
     */
    public Vector getObComponentByBaseName(String name) {
        Vector rc = (Vector) indexes[BASE_NAME_INDEX].get(name);
        return rc;
    }

    /**
     * @param name The base name of the OBComponents to be located.
     * @param dirName The name of the directory the file must be in.
     * @return Finding an OBComponent by its base name will return a
     * Vector of components (as multiple components can share
     * the same base name) or null if none are located.
     * Note: an empty Vector will never be returned, but elements
     * in the returned Vector may be null.
     * This method uses the FULL CANONICAL pathname to files. That
     * is: all links and relative paths have been resolved. This is done
     * easily with File.getCanonicalPath().
     */
    public Vector getObComponentByBaseName(String name, String dirName) {
        Vector baseMatch = getObComponentByBaseName(name);
        Vector rc = new Vector();
        if (baseMatch != null) {
            dirName = DFSFile.getCanonicalPath(dirName);
            if (dirName != null) {
                Enumeration e = baseMatch.elements();
                while (e.hasMoreElements()) {
                    OBComponent obc = (OBComponent) e.nextElement();
                    String directory =
                        obc.getSourceFile().getPath().getParent();
                    if (directory.equals(dirName)) {
                        rc.addElement(obc);
                    }
                }
            }
        }
        return rc;
    }

    /**
     * @param name The id name of the OBComponent to be located.
     * @return The single OBComponent located (or null for a failed search).
     * This method uses the FULL CANONICAL pathname to files. That
     * is: all links and relative paths have been resolved. This is done
     * easily with File.getCanonicalPath().
     */
    public OBComponent getObComponentByName(String name) {
        OBComponent rc = null;
        name = DFSFile.getCanonicalPath(name);
        if (name != null) {
            rc = (OBComponent) indexes[NAME_INDEX].get(name);
        }
        return rc;
    }

    /**
     * @param name The id name of the OBComponent to be located.
     * @return The single OBComponent located (or null for a failed search).
     * This method *does not* use full canonical pathnames
     */
    public OBComponent getObComponentByFakeName(String name) {
        OBComponent rc = null;
        if (name != null) {
            rc = (OBComponent) indexes[NAME_INDEX].get(name);
        }
        return rc;
    }

    /**
     * @param name The id name of the OBComponent to be located.
     * @return The single OBComponent located (or null for a failed search).
     * This method *does not* use full canonical pathnames
     */
    public OBComponent getObComponentBySignature(String sig) {
        OBComponent rc = null;
        if (sig != null) {
            rc = (OBComponent) indexes[MD5_INDEX].get(sig);
        }
        return rc;
    }

    /**
     * @param name The arcfile name of the OBComponents to be located.
     * @return OBComponent located, or null if not found.
     */

    public OBComponent getObComponentByArcname(String name) {
        OBComponent rc = null;
        String backName = new String(name);

        name = DFSFile.getCanonicalPath(name);
        if (name != null) {
            if (useArcNameIndex) {
                rc = (OBComponent) indexes[ARC_NAME_INDEX].get(name);
                // if for whatever reason there was no ARCFILE in the index,
                // return the normal name
                if (rc == null) {
                    return getObComponentByName(backName);
                }
            } else {
                return getObComponentByName(backName);
            }
        }
        return rc;
    }

    /**
     * @param obc The OBComponent to be added to the indexes.
     * Add an OBComponent to the various indexes.
     * There is no check in this method to ensure obc was not already added.
     */
    private void addObComponentToIndexes(OBComponent obc) {

        // add to the name index.
        indexes[NAME_INDEX].put(obc.getId(), obc);
        // add the arcfile index
        if (useArcNameIndex
            && (obc instanceof FitsFileOBComponent)
            && !obc.isPipeProduct()) {
            FitsFileOBComponent ffoc = (FitsFileOBComponent) obc;
            indexes[ARC_NAME_INDEX].put(ffoc.getArcfile(), obc);
        }

        // add to the Filesystem base name index.
        indexes[FS_BASE_NAME_INDEX].put(DFSFile.getFileName(obc.getId()), obc);

        // add to the base name index.
        String baseName = DFSFile.extractBaseName(obc.getId());
        Vector v = getObComponentByBaseName(baseName);
        if (v == null) {
            v = new Vector();
            indexes[BASE_NAME_INDEX].put(baseName, v);
        }
        v.addElement(obc);

        // add to the MD5 signature index
        if (obc instanceof KeywordContainerOBComponent) {
            String signature =
                ((KeywordContainerOBComponent) obc).getSignature();
            if (!signature.equals("")) {
                indexes[MD5_INDEX].put(signature, obc);
                //System.out.println("Added signature "+signature+" for component "+obc);
            }
        }
    }

    /**
     * Remove a component from the various indexes.
     * @param obc The OBComponent to be removed from the indexes.
     */
    private void removeObComponentFromIndexes(OBComponent obc) {
        // remove obc from the name index.
        indexes[NAME_INDEX].remove(obc.getId());
        // remove obc from the arcfile index.
        if (useArcNameIndex
            && (obc instanceof FitsFileOBComponent)
            && !obc.isPipeProduct()) {
            FitsFileOBComponent ffoc = (FitsFileOBComponent) obc;
            indexes[ARC_NAME_INDEX].remove(ffoc.getArcfile());
        }
        // remove from  the Filesystem base name index.
        indexes[FS_BASE_NAME_INDEX].remove(DFSFile.getFileName(obc.getId()));

        // remove obc from the base name index.
        String baseName = DFSFile.extractBaseName(obc.getId());
        Vector v = getObComponentByBaseName(baseName);
        if (v != null) {
            v.removeElement(obc);
            if (v.isEmpty()) {
                indexes[BASE_NAME_INDEX].remove(baseName);
            }
        }
    }

    /**
     * @return The SortedHashtable of Program objects in this DataModel.
     */
    public SortedHashtable getProgramSet() {
        return programs;
    }

    public SortedHashtable getTelescopesSet() {
        return telescopes;
    }

    /**
     * Get the SortedHashtable for the requested directory.
     * @param dir
     * @return SortedHashtable
     */
    public SortedHashtable getDirectory(String dir) {
        return (SortedHashtable) directories.get(dir);
    }

    /**
     * Get the element at index i of the directories sorted hashtable.
     * @param i index required, starting at 0.
     * @return SortedHashtable at index i.
     */
    public SortedHashtable getDirectoryAt(int i) {
        return (SortedHashtable) directories.elementAt(i);
    }

    /**
     * Get the index of a directory in the directories hashtable.
     * @param dir
     * @return int the index of the dir in the directories dictionary.
     */
    public int getDirectoryIndex(Object dir) {
        return directories.indexOf(dir);
    }

    /**
     * Get the number of entries in the directories sorted hashtable.
     * @return int the number of directories.
     */
    public int getDirectoryCount() {
        return directories.size();
    }

    /**
     * Get an enumeration of the directories stored.
     * @return Enumeration of directories
     */
    public Enumeration getDirectories() {
        return directories.elements();
    }

    public SortedHashtable getTelescope(String tel) {
        return (SortedHashtable) telescopes.get(tel);
    }

    /**
     * @param programId The id of the Program for which the set of
     * observations is required.
     * @return The SortedHashtable of observations for the program id specified,
     * or null if there is no such program id in this DataModel.
     */
    public SortedHashtable getObservationSet(String programId) {
        return getObservationSet(programs, programId);
    }

    private SortedHashtable getObservationSet(
        SortedHashtable progSet,
        String programId) {
        SortedHashtable rc = null;
        Program p = (Program) progSet.get(programId);
        if (p != null) {
            rc = p.getObservationSet();
        }
        return rc;
    }

    /**
     * Get the NAME_INDEX index.
     * @return Hashtable
     */
    public Hashtable getFilesByName() {
        return indexes[NAME_INDEX];
    }

    /**
     * Get the number of files in the model.
     * Used by the DataModelBuilder dialog to show update() progress.
     */
    public int getTotalFilesInSet() {
        return totalFilesInSet;
    }

    /**
     * Get the number of files updated.
     * Used by the DataModelBuilder dialog to show update() progress.
     */
    public int getNumberOfFilesUpdated() {
        return numberOfFilesUpdated;
    }

    /**
     * Get all files from all file selection models.
     * @return A Vector of Vector of file names.
     * One Vector for each fileSelectionModel.
     * Notice that this method will trigger
     * a re-read of all directories.
     */
    public Vector getFileSet() {
        Vector rc = new Vector(fileSelectionModels.size());
        Enumeration fileSelections = fileSelectionModels.elements();
        int total = 0;
        while (fileSelections.hasMoreElements()) {
            FileSelectionModel fileSelection =
                (FileSelectionModel) fileSelections.nextElement();
            Vector fileList = fileSelection.getFileList();
            total += fileList.size();
            rc.addElement(fileList);
        }
        totalFilesInSet = total;
        numberOfFilesUpdated = 0;
        return rc;
    }

    /**
     * Update the model: re-read the list of files from the
     * FileSelectionModels, update the files in the model and
     * fire any DataModelChangeEvents.
     * Warning: calling update will also apply any new filters and
     * RuleTables which have been set.
     */
    public void update() {
        update(getFileSet());
    }

    /**
     * Update the model with the vector of file vectors.
     * @param vectorOfFileVectors
     */
    public synchronized void update(Vector vectorOfFileVectors) {
        // 1. Make a shallow copy of the name index.
        // 2. Enumerate through each file in each file selection model.
        // 3.   Search for each component:
        // 4.     if located, remove from the index copy and add to 
        //          the list of components to have update() called.
        // 5.     else addFile().
        // 6. Remove from this DataModel all components 
        //    remaining in the copy index.
        // 7. Enumerate through the index calling update() on each
        //    component.

        DataModelEvent modelChangedEvent = null;

        // indicate that the model is updating.
        dataModelUpdating = true;

        // Hashtable.clone() is expensive - is there a better way
        // to do this?
        Hashtable tempIndex = (Hashtable) indexes[NAME_INDEX].clone();

        Vector updateSet = new Vector();

        // enumerate through each file of each FileSelectionModel.
        Enumeration fileLists = vectorOfFileVectors.elements();
        while (fileLists.hasMoreElements()) {
            Vector fileList = (Vector) fileLists.nextElement();
            Enumeration files = fileList.elements();
            while (files.hasMoreElements()) {
                String fileName = (String) files.nextElement();
                numberOfFilesUpdated++;
                OBComponent comp = (OBComponent) tempIndex.remove(fileName);
                if (comp == null) {
                    // the file did not previously exist - i.e. it is new in the model.
                    int rc = addFile(fileName);
                    if (rc != DFSFile.FILE_VALID
                        && rc != DFSFile.FILE_FILTERED) {
                        // do not print bothering messages if debug lower than 1
                        if (PropertyDB
                            .getInstance()
                            .isDebug(PropertyDB.debugDataModelChange)) {
                            System.out.println(
                                ""
                                    + fileName
                                    + ": "
                                    + DFSFile.getStringStatus(rc)
                                    + ", ignored.");
                        }
                    } else {
                        // file has been already added to the model ***********
                        modelChangedEvent =
                            new DataModelEvent(
                                DataModelEvent.FILE_ADDED,
                                getObComponentByName(fileName));
                        fireDataModelChangeEvent(modelChangedEvent);
                    }
                } else {
                    // it is in this DataModel, so we must update it.
                    updateSet.addElement(comp);
                }
            }
        }

        // anything left in the tempIndex was not a part of the file set
        // and so no longer exists. i.e. must be removed from the model.
        Enumeration compToRemove = tempIndex.elements();
        int eCount = tempIndex.size();
        while (compToRemove.hasMoreElements()) {
            OBComponent c = (OBComponent) compToRemove.nextElement();
            // first we send the Event, only if we are deleting limited numbers
            // of files. If it's an entire directory deletion we avoid wasting time
            if (eCount < eventTolerance) {
                modelChangedEvent =
                    new DataModelEvent(DataModelEvent.FILE_DELETED, c);
                fireDataModelChangeEvent(modelChangedEvent);
            } else {
                removeObComponent(c);
            }

        }
        //
        // all existing elements that were not removed need to be updated.
        //
        Enumeration updateList = updateSet.elements();
        while (updateList.hasMoreElements()) {
            OBComponent c = (OBComponent) updateList.nextElement();
            c.update();
        }

        tempIndex.clear();
        tempIndex = null;
        dataModelUpdating = false;
    }

    /**
     * tentative for a semaphor
     * @returns: a semaphor to indicate that the
     * update method is still running
     *
     */
    public boolean isUpdating() {
        return dataModelUpdating;
    }

    /**
     * @param fileSelection The FileSelectionModel of files to be
     * updated.
     * @param refreshFiles Call update() on the components or not.
     */
    /*
    private void addFiles(FileSelectionModel fileSelection) {
        Vector fileList = fileSelection.getFileList();
        Enumeration files = fileList.elements();
        while (files.hasMoreElements()) {
            String fileName = (String) files.nextElement();
            int rc = addFile(fileName);
            if (rc != DFSFile.FILE_VALID) {
                System.out.println(
                    "Error: "
                        + fileName
                        + ": "
                        + DFSFile.getStringStatus(rc)
                        + ", ignored.");
            }
        }
    }
    */

    /**
     * @param fileName The name of the file to be added to this DataModel.
     * Add a file to the Model. Determine the file type (from the
     * extension and then create the appropriate OBComponent for the file.
     * If the OBComponent created is not valid, it is deleted at the
     * and of the method.
     * 
     * The addOBComponent() call to this DataModel is made by each class
     * of OBComponent during construction/building of the component.
     * 
     * @return The status of the component created.
     * @see org.eso.gasgano.DFSFile
     */
    private int addFile(String fileName) {
        int rc = DFSFile.FILE_INVALID;
        OBComponent comp = null;
        if (fileName.endsWith(".fits")
            || fileName.endsWith(".fit")
            || fileName.endsWith(".FIT")
            || fileName.endsWith(".fits.Z")
            || fileName.endsWith(".fits.gz")) {
            comp = new FitsFileOBComponent(fileName);
        } else if (fileName.endsWith(".tfits")) {
            comp = new TFitsFileOBComponent(fileName);
        } else if (
            fileName.endsWith(".rX")
                || fileName.endsWith(".rY")
                || fileName.endsWith(".rZ")
                || fileName.endsWith(".rb")) {
            comp = new ReductionBlockOBComponent(fileName);
        } else {
            if (PropertyDB.getInstance().getProperty("TEXTFILE_EXTENSIONS")
                != null) {
                Vector extTmp =
                    GasProp.stringScanner(
                        PropertyDB.getInstance().getProperty(
                            "TEXTFILE_EXTENSIONS"),
                        ';');
                for (int i = 0; i < extTmp.size(); i++) {
                    if (fileName.endsWith((String) extTmp.elementAt(i))) {
                        comp = new DFSTextFileOBComponent(fileName);
                        break;
                    }
                }
            }
        }
        if (comp == null
            && PropertyDB.getInstance().isDebug(
                PropertyDB.debugDataModelChange)) {
            System.out.println(fileName + ": not recognized");
        }

        if (comp != null) {
            KeywordContainerOBComponent tmpHeader = null;
            if (comp instanceof KeywordContainerOBComponent) {
                tmpHeader = (KeywordContainerOBComponent) comp;
                if (tmpHeader.containsRbInfo()) {
                    // this is the only place where VRB are instantiated
                    if (getAssociatedVRB(DFSFile
                        .extractBaseName(tmpHeader.getId()))
                        == null) {
                        VirtualReductionBlockOBComponent vrb =
                            new VirtualReductionBlockOBComponent(tmpHeader);
                        comp.addAssociatedReductionBlock(vrb);
                        addDataModelChangeListener(vrb);
                        addAssociatedVRB(
                            DFSFile.extractBaseName(tmpHeader.getId()),
                            vrb);
                    }
                }
                tmpHeader.freeHeaderMemory();
            }
            rc = comp.getStatus();
        }

        if (rc == DFSFile.FILE_VALID
            && PropertyDB.getInstance().isDebug(
                PropertyDB.debugDataModelChange)) {
            System.out.println(fileName + ": added.");
        }
        return rc;
    }

    /**
     * Reclassify all FitsFiles in the data model.
     */
    public void reclassify() {
        Enumeration files = indexes[NAME_INDEX].elements();
        long startTime = 0;
        int filesClassified = 0;
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis();
        }
        while (files.hasMoreElements()) {
            OBComponent obc = (OBComponent) files.nextElement();
            if (obc instanceof FitsFileOBComponent) {
                FitsFileOBComponent ffobc = (FitsFileOBComponent) obc;
                ffobc.classify();
                filesClassified++;
            }
        }
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis() - startTime;
            System.out.println(
                "Reclassified "
                    + filesClassified
                    + " files in "
                    + startTime
                    + " milliseconds.");
        }
    }

    public void resetDescription() {
        Enumeration files = indexes[NAME_INDEX].elements();
        long startTime = 0;
        int filesReset = 0;
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis();
        }
        while (files.hasMoreElements()) {
            OBComponent obc = (OBComponent) files.nextElement();
            if (obc instanceof KeywordContainerOBComponent) {
                KeywordContainerOBComponent hhobc =
                    (KeywordContainerOBComponent) obc;
                hhobc.resetDescription();
                filesReset++;
            }
        }
        if (PropertyDB.getInstance().isDebug(PropertyDB.debugTiming)) {
            startTime = System.currentTimeMillis() - startTime;
            System.out.println(
                "Description Reset "
                    + filesReset
                    + " files in "
                    + startTime
                    + " milliseconds.");
        }
    }

    /**
     * @param rt The RuleTable to be used to classify FitsFiles in this
     * DataModel. Setting a ruleTable will not apply the rules to the 
     * files in the model - this must be done with reclassify() or update().
     */
    public void setRuleTable(RuleTable rt) {
        ruleTable = rt;
    }

    /**
     * @return The RuleTable used to classify FitsFiles in this DataModel.
     */
    public RuleTable getRuleTable() {
        return ruleTable;
    }

    /**
     * @param f The OBComponentFilter used to filter files in this DataModel.
     * Setting a filter will not apply it - this must be
     * done by calling applyFilter() or update().
     */
    public void setFileFilter(OBComponentFilter f) {
        fileFilter = f;
    }

    /**
     * @return The OBComponentFilter used to filter files in this DataModel.
     */
    public OBComponentFilter getFileFilter() {
        return fileFilter;
    }

    public void applyFilter() {
        Enumeration files = indexes[NAME_INDEX].elements();
        DataModelEvent modelChangedEvent = null;
        while (files.hasMoreElements()) {
            OBComponent obc = (OBComponent) files.nextElement();
            // We want to filter all OBComponents, not just Keyword
            // containers - this is reported in bug DFS01236.
            // Original code:
            // if (obc instanceof KeywordContainerOBComponent
            //    && !passFilter(obc)) {
            if (!passFilter(obc)) {
                // removeObComponent( obc);
                // above line now handled by the Listener
                modelChangedEvent =
                    new DataModelEvent(DataModelEvent.FILE_DELETED, obc);
                fireDataModelChangeEvent(modelChangedEvent);
            }
        }
    }

    /**
     * determine if an OBComponent passes the current filter, if set.
     * @param comp the OBComponent 
     * @return boolean true if there is no filter or it passes the filter
     */
    public boolean passFilter(OBComponent comp) {
        boolean rc = true;
        if (fileFilter != null) {
            rc = fileFilter.accept(comp);
        }
        return rc;
    }

    public SortedHashtable addTelescope(String telName) {
        DataModelEvent modelChangedEvent = null;
        SortedHashtable rc = findTelescope(telName);
        if (rc == null) {
            rc = new SortedHashtable(telName);
            telescopes.put(telName, rc);
            modelChangedEvent =
                new DataModelEvent(DataModelEvent.TELESCOPE_ADDED, rc);
            modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
            fireDataModelChangeEvent(modelChangedEvent);
        }
        if (PropertyDB
            .getInstance()
            .isDebug(PropertyDB.debugDataModelChange)) {
            System.out.println("Added Telescope: " + telName);
        }
        return rc;
    }

    public SortedHashtable findTelescope(String telName) {
        return (SortedHashtable) telescopes.get(telName);
    }

    public void removeTelescope(String telName) {
        DataModelEvent modelChangedEvent = null;
        modelChangedEvent =
            new DataModelEvent(
                DataModelEvent.TELESCOPE_DELETED,
                findTelescope(telName));
        modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
        telescopes.remove(telName);
        fireDataModelChangeEvent(modelChangedEvent);
        if (PropertyDB
            .getInstance()
            .isDebug(PropertyDB.debugDataModelChange)) {
            System.out.println("Removed Telescope: " + telName);
        }
    }

    public SortedHashtable addDirectory(String dirName) {
        DataModelEvent modelChangedEvent = null;
        SortedHashtable rc = findDirectory(dirName);

        if (rc == null) {
            rc = new SortedHashtable(dirName);
            directories.put(dirName, rc);
            modelChangedEvent =
                new DataModelEvent(
                    DataModelEvent.DIRECTORY_ADDED,
                    findDirectory(dirName));
            modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
            fireDataModelChangeEvent(modelChangedEvent);
        }
        return rc;
    }

    public SortedHashtable findDirectory(String dirName) {
        return (SortedHashtable) directories.get(dirName);
    }

    public void removeDirectory(String dirName) {
        DataModelEvent modelChangedEvent = null;
        modelChangedEvent =
            new DataModelEvent(
                DataModelEvent.DIRECTORY_DELETED,
                findDirectory(dirName));
        modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
        directories.remove(dirName);
        fireDataModelChangeEvent(modelChangedEvent);
    }

    public Program addProgram(String id, OBComponent obc) {
        return addProgram(programs, id, obc);
    }

    /**
        * @param id The id of the Program to be added.
        * @param obc The OBComponent which requires the Program added.
        * @return The Program added to this DataModel.
        */
    private Program addProgram(
        SortedHashtable progSet,
        String id,
        OBComponent obc) {
        DataModelEvent modelChangedEvent = null;
        Program rc = findProgram(progSet, id);
        if (rc == null) {
            rc = new Program(id);

            if (progDescFactory != null) {
                progDescFactory.setDescription(rc, obc);
            }

            progSet.put(id, rc);
            modelChangedEvent =
                new DataModelEvent(DataModelEvent.PROGRAM_ADDED, rc);
            if (progSet == programs) {

                modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
            } else if (findDirectory(progSet.toString()) != null) {
                rc.setDirectory(progSet);
                modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
            } else if (findTelescope(progSet.toString()) != null) {
                rc.setTelescope(progSet);
                modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
            } else {
                System.out.println("Unidentified progSet " + progSet);
            }

            fireDataModelChangeEvent(modelChangedEvent);

            if (PropertyDB
                .getInstance()
                .isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println(
                    "Added Prog: "
                        + id
                        + " to "
                        + ((progSet == programs)
                            ? "programs"
                            : progSet.toString()));
                // displayHashKeys(progSet);
            }
        }
        return rc;
    }

    /**
     * @param id The id of the Program to be located.
     * @return The Program located, or null if the Program does not exist
     * in this DataModel.
     */
    public Program findProgram(String id) {
        return findProgram(programs, id);
    }

    /**
     * Find the program given the program set and id.
     * @param progSet
     * @param id
     * @return Program
     */
    public Program findProgram(SortedHashtable progSet, String id) {
        return (Program) progSet.get(id);
    }

    /**
     * @param id The id of the Program to be removed from this DataModel.
     */
    public void removeProgram(String id) {
        removeProgram(programs, id);
    }

    private void removeProgram(SortedHashtable progSet, String id) {
        DataModelEvent modelChangedEvent = null;
        Program tmpProgram = null;
        tmpProgram = findProgram(progSet, id);
        if (tmpProgram != null) {
            modelChangedEvent =
                new DataModelEvent(DataModelEvent.PROGRAM_DELETED, tmpProgram);
            if (progSet == programs) {

                modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
            } else if (findDirectory(progSet.toString()) != null) {

                modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
            } else if (findTelescope(progSet.toString()) != null) {

                modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
            } else {
                System.out.println(
                    "DataModel::removeProgram: Unidentified progSet "
                        + progSet);
            }

            progSet.remove(id);
            fireDataModelChangeEvent(modelChangedEvent);

        }

        if (PropertyDB
            .getInstance()
            .isDebug(PropertyDB.debugDataModelChange)) {
            System.out.println(
                "Removed Prog: "
                    + id
                    + " from "
                    + ((progSet == programs) ? "programs" : progSet.toString()));
        }
    }

    public Observation addObservationAndFire(
        Program prog,
        String obId,
        OBComponent obc) {
        DataModelEvent modelChangedEvent = null;
        Observation obs = null;

        if (prog.findObservation(obId) == null) {
            obs = prog.addObservation(obId, obc);
            obs.setProgram(prog);
            modelChangedEvent =
                new DataModelEvent(DataModelEvent.OBSERVATION_ADDED, obs);
            if (prog.getTelescope() != null) {
                // it's an observation for a program for the default view
                modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
            } else if (prog.getDirectory() != null) {
                // it's an observation for a program for the directory view
                modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
            } else if (findProgram(prog.getId()) != null) {
                // it's an observation for a program for the telescope  view
                modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
            } else {
                System.out.println(
                    "DataModel::addObservation: Unidentified progSet " + prog);
                displayHashKeys(programs);
                System.out.flush();
                System.out.println("%%%%%%%%%%%%%%%%");
                return obs;
            }

            fireDataModelChangeEvent(modelChangedEvent);
        } else {
            obs = prog.addObservation(obId, obc);
        }
        return obs;
    }

    /**
     * Get the named intermediate (group) from the given Observation, creating
     * the intermediate if it does not already exist. Events are fired to data model
     * listeners as required.
     * @param obs
     * @param name
     * @return InterContainer
     */
    public InterContainer addIntermediate(Observation obs, OBComponent obc, int treeModelMode, String name) {
        InterContainer rc = null;
        DataModelEvent modelChangedEvent = null;
        rc = obs.findIntermediate(name);
        if (rc == null) {
            rc = obs.addIntermediate(name);
            modelChangedEvent =
                new DataModelEvent(DataModelEvent.INTERMEDIATE_ADDED, rc);
            Program prog = obs.getProgram();
            if (prog.getTelescope() != null) {
                // it's an intermediate for a program for the default view
                modelChangedEvent.setTreeType(GasProp.TREE_TELESCOPE);
            } else if (prog.getDirectory() != null) {
                // it's an intermediate for a program for the directory view
                modelChangedEvent.setTreeType(GasProp.TREE_DIRECTORY);
            } else if (findProgram(prog.getId()) != null) {
                // it's an observation for a program for the telescope  view
                modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
            } else {
                System.out.println(
                    "DataModel::addIntermediate: Unidentified progSet " + prog);
                displayHashKeys(programs);
                System.out.flush();
                System.out.println("%%%%%%%%%%%%%%%%");
                // this is an error state - should not happen
                // return without completing structure or firing events.
                return rc;
            }
        }
        rc.addObComponent(obc);
        obc.setObservation(rc, treeModelMode);
        rc.setObservation(obs);
        if (modelChangedEvent != null) {
            fireDataModelChangeEvent(modelChangedEvent);
        }
        return rc;
    }
    
    /**
     * Remove the given intermediate from the given observation and fire an event to
     * notify listeners.
     * @param obs
     * @param inter
     * @param treeModelMode
     */
    public void removeIntermediate(Observation obs, InterContainer inter, int treeModelMode) {
        DataModelEvent modelChangedEvent =
            new DataModelEvent(
                DataModelEvent.INTERMEDIATE_DELETED,
                inter);
        modelChangedEvent.setTreeType(treeModelMode);        
        obs.removeIntermediate(inter);
        inter.cleanup();
        fireDataModelChangeEvent(modelChangedEvent);
    }
    
    /**
     * @param progId The id of the Program under which the OBComponent
     * belongs.
     * @param obId The id of the observation under which ths OBComponent
     * belongs.
     * @param obc The OBComponent to be added to this DataModel.
     * 
     * Add an OBComponent to the DataModel. The Program and Observations
     * will be updated to include the OBComponent, and the program, observation
     * and date will be set in the OBComponent.
     * It is important that an OBComponent be added only once, to add it
     * again under another program, observation or date, it must first 
     * be removed using removeComponent().
     * A component will not be added if there is already a component of the
     * same name in the NAME_INDEX index.
     */
    public void addObComponent(
        String progId,
        String obId,
        String directoryName,
        String telName,
        OBComponent obc) {
        boolean debug = false;
        boolean interMediate = false;
        String interMediateName = null;

        if (getObComponentByName(obc.getId()) == null) {

            // determine whether intermediate container will be needed
            if (obc instanceof KeywordContainerOBComponent) {
                interMediate =
                    ((KeywordContainerOBComponent) obc).hasIntermediate();
                interMediateName = obc.getIntermediateGrouping();
            }

            // add the component to the indexes.
            addObComponentToIndexes(obc);

            if (debug) {
                System.out.println("********* Prog/OB Tree");
                System.out.flush();
            }

            // add component to Prog/ob tree.
            Program prog = addProgram(progId, obc);
            Observation obs = addObservationAndFire(prog, obId, obc);
            if (interMediate) {
                InterContainer inter = addIntermediate(obs, obc, GasProp.TREE_DEFAULT, interMediateName);
            } else {
                obs.addObComponent(obc);
                // update the component with its prog and obs id.
                obc.setObservation(obs, GasProp.TREE_DEFAULT);
            }

            if (debug) {
                System.out.println("********* Directory/ Prog/OB Tree");
                System.out.flush();
            }

            // add component to directory/prog/ob tree.
            SortedHashtable progSet = addDirectory(directoryName);
            prog = addProgram(progSet, progId, obc);
            obs = addObservationAndFire(prog, obId, obc);
            if (interMediate) {
                InterContainer inter = addIntermediate(obs, obc,
                    GasProp.TREE_DIRECTORY, interMediateName);
            } else {
                obs.addObComponent(obc);
                obc.setObservation(obs, GasProp.TREE_DIRECTORY);
            }

            if (debug) {
                System.out.println("********* Telescope /Prog /OB Tree");
                System.out.flush();
            }

            // add component to telescope/prog/ob tree.
            progSet = addTelescope(telName);
            prog = addProgram(progSet, progId, obc);
            obs = addObservationAndFire(prog, obId, obc);

            if (interMediate) {
                InterContainer inter = addIntermediate(obs, obc, 
                    GasProp.TREE_TELESCOPE, interMediateName);
            } else {
                obs.addObComponent(obc);
                obc.setObservation(obs, GasProp.TREE_TELESCOPE);
            }

            // System.err.println(telName+" => "+progId+" => "+obId+" => "+obc);
            if (PropertyDB
                .getInstance()
                .isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println(
                    "Added: " + obc.getId() + " to: " + progId + " " + obId);
            }
        }
    }

    /**
     * @param name The name of the file to be removed from the DataModel.
     */
    public void removeObComponent(String name) {
        OBComponent obc = getObComponentByName(name);
        removeObComponent(obc);
    }

    /**
     * method used merely for debugging purposes
     * prints on stderr the files of programs and their corresponding
     * observations, with their amount of obcomponents
     */
    public void obsProgSummary() {
        Enumeration progE = null;
        Enumeration obsE = null;
        Enumeration dirE = directories.elements();
        while (dirE.hasMoreElements()) {
            SortedHashtable dir = (SortedHashtable) dirE.nextElement();
            System.out.println(dir + ":");
            progE = dir.elements();
            while (progE.hasMoreElements()) {
                Program prog = (Program) progE.nextElement();
                System.out.println("\t" + prog);
                obsE = prog.getObservationSet().elements();
                while (obsE.hasMoreElements()) {
                    Observation obs = (Observation) obsE.nextElement();
                    System.out.println(
                        "\t\t<"
                            + obs
                            + "> "
                            + obs.getComponentCount()
                            + " components");
                }
            }
        }
    }

    /**
     * @param obc The OBComponent to be removed from this DataModel.
     * 
     * Remove an ObComponent, if located. The OBComponent will have its
     * program, observation and obdate set to null once removed.
     * Any observations, obdates or programs which become empty 
     * after removal of this ObComponent, will be removed from the model.
     * The OBComponent is removed from the name and base name index as well.
     */
    public void removeObComponent(OBComponent obc) {

        Observation obs = null;
        InterContainer inter = null;
        Program prog = null;
        SortedHashtable telescope = null;
        SortedHashtable directory = null;
        boolean interMediate = false;

        DataModelEvent modelChangedEvent = null;
        int delay = 0;

        // remove from indexes.
        removeObComponentFromIndexes(obc);

        if (obc instanceof KeywordContainerOBComponent) {
            interMediate =
                ((KeywordContainerOBComponent) obc).hasIntermediate();
        }

        // remove from dir/prog/ob tree

        if (interMediate) {
            inter = (InterContainer) obc.getObservation(GasProp.TREE_DIRECTORY);
            obs = inter.getObservation();
        } else {
            obs = obc.getObservation(GasProp.TREE_DIRECTORY);
        }
        prog = obs.getProgram();
        directory = prog.getDirectory();

        if (directory != null) {

            if (prog != null) {

                if (obs != null) {
                    if (interMediate) {
                        if (inter != null) {
                            inter.removeObComponent(obc);
                            if (inter.isEmpty()) {
                                removeIntermediate(obs, inter, GasProp.TREE_DIRECTORY);
                            }
                            if (obs.isEmpty()) {
                                modelChangedEvent =
                                    new DataModelEvent(
                                        DataModelEvent.OBSERVATION_DELETED,
                                        obs);
                                modelChangedEvent.setTreeType(
                                    GasProp.TREE_DIRECTORY);
                                prog.removeObservation(obs.getId());
                                obs.cleanup();
                                fireDataModelChangeEvent(modelChangedEvent);
                            }

                        }
                    } else {
                        obs.removeObComponent(obc);
                        if (obs.isEmpty()) {
                            modelChangedEvent =
                                new DataModelEvent(
                                    DataModelEvent.OBSERVATION_DELETED,
                                    obs);
                            modelChangedEvent.setTreeType(
                                GasProp.TREE_DIRECTORY);
                            prog.removeObservation(obs.getId());
                            obs.cleanup();
                            fireDataModelChangeEvent(modelChangedEvent);
                        }
                    }
                }
                if (prog.isEmpty()) {
                    sleep(delay);
                    removeProgram(directory, prog.getId());
                    prog.cleanup();
                }
            }
            if (directory.isEmpty()) {
                removeDirectory(directory.toString());
            }
        }

        // remove from telescope/prog/ob tree
        if (interMediate) {
            inter = (InterContainer) obc.getObservation(GasProp.TREE_TELESCOPE);
            obs = inter.getObservation();
        } else {
            obs = obc.getObservation(GasProp.TREE_TELESCOPE);
        }
        prog = obs.getProgram();
        telescope = prog.getTelescope();

        if (telescope != null) {
            if (prog != null) {
                if (obs != null) {
                    if (interMediate) {
                        if (inter != null) {
                            inter.removeObComponent(obc);
                            if (inter.isEmpty()) {
                                removeIntermediate(obs, inter, GasProp.TREE_TELESCOPE);
                            }
                            if (obs.isEmpty()) {
                                modelChangedEvent =
                                    new DataModelEvent(
                                        DataModelEvent.OBSERVATION_DELETED,
                                        obs);
                                modelChangedEvent.setTreeType(
                                    GasProp.TREE_TELESCOPE);
                                prog.removeObservation(obs.getId());
                                obs.cleanup();
                                fireDataModelChangeEvent(modelChangedEvent);
                            }

                        }
                    } else {
                        obs.removeObComponent(obc);
                        if (obs.isEmpty()) {
                            modelChangedEvent =
                                new DataModelEvent(
                                    DataModelEvent.OBSERVATION_DELETED,
                                    obs);
                            modelChangedEvent.setTreeType(
                                GasProp.TREE_TELESCOPE);
                            prog.removeObservation(obs.getId());
                            obs.cleanup();
                            fireDataModelChangeEvent(modelChangedEvent);
                        }
                    }
                }
                if (prog.isEmpty()) {
                    prog.cleanup();
                    sleep(delay);
                    removeProgram(telescope, prog.getId());
                }
            }
            if (telescope.isEmpty()) {
                removeTelescope(telescope.toString());
            }
        }

        // remove from program/ob tree.
        if (interMediate) {
            inter = (InterContainer) obc.getObservation(GasProp.TREE_DEFAULT);
            obs = inter.getObservation();
        } else {
            obs = obc.getObservation(GasProp.TREE_DEFAULT);
        }
        prog = obs.getProgram();

        if (prog != null) {
            if (obs != null) {
                if (interMediate) {
                    if (inter != null) {
                        inter.removeObComponent(obc);
                        if (inter.isEmpty()) {
                            removeIntermediate(obs, inter, GasProp.TREE_DEFAULT);
                        }
                        if (obs.isEmpty()) {
                            modelChangedEvent =
                                new DataModelEvent(
                                    DataModelEvent.OBSERVATION_DELETED,
                                    obs);
                            modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
                            prog.removeObservation(obs.getId());
                            fireDataModelChangeEvent(modelChangedEvent);
                        }
                    }
                } else {
                    obs.removeObComponent(obc);
                    if (obs.isEmpty()) {
                        modelChangedEvent =
                            new DataModelEvent(
                                DataModelEvent.OBSERVATION_DELETED,
                                obs);
                        modelChangedEvent.setTreeType(GasProp.TREE_DEFAULT);
                        prog.removeObservation(obs.getId());
                        fireDataModelChangeEvent(modelChangedEvent);
                    }
                }
            }
            if (prog.isEmpty()) {
                sleep(delay);
                removeProgram(prog.getId());
            }

            // execute the necessary closing of the OBComponent.
            obc.cleanup();

            if (PropertyDB
                .getInstance()
                .isDebug(PropertyDB.debugDataModelChange)) {
                System.out.println("Removed: " + obc.getId());
            }
        }
    }

    /**
     * utility method required by the model - sleep for a small delay.
     * @param x
     */
    private static void sleep(int x) {
        try {
            Thread.sleep(x);
        } catch (InterruptedException ie) {
        }
    }

    /**
     * Get the program at the given index.
     * @param i
     * @return Program
     */
    public Program getProgramAt(int i) {
        return (Program) programs.elementAt(i);
    }

    /**
     * get the index of a given program in the index.
     * @param prog
     * @return int
     */
    public int getIndexOfProgram(Program prog) {
        return programs.indexOf(prog);
    }

    /**
     * Get the number of programs in the model. 
     * @return int
     */
    public int getProgramCount() {
        return programs.size();
    }

    /**
     * Query the DataModel for all associated files based on the file
     * passed in the method.
     * The Vector parameters are filled with String objects contains
     * file names.
     */
    public void getAssociatedFiles(
        OBComponent obc,
        Vector inputFiles,
        Vector masterCalibFiles,
        Vector reductionBlocks,
        Vector productFiles) {
        Vector associatedRbs = obc.getAssociatedReductionBlocks();
        if (associatedRbs != null) {
            Enumeration e = associatedRbs.elements();
            while (e.hasMoreElements()) {
                ReductionBlockCommon rbComp =
                    (ReductionBlockCommon) e.nextElement();
                // input files list.
                copyVector(rbComp.getInputFileNames(), inputFiles);
                // master calib file list.
                copyVector(rbComp.getMasterCalibFileNames(), masterCalibFiles);
                // reduction block list
                reductionBlocks.addElement(rbComp.getId());
                // product file list.
                copyVector(rbComp.getProductFileNames(), productFiles);
            }
        }
    }

    /**
     * Get a virtual reduction block for the given base product name.
     */
    public VirtualReductionBlockOBComponent getAssociatedVRB(String productBaseName) {
        VirtualReductionBlockOBComponent rc = null;
        rc =
            (VirtualReductionBlockOBComponent) indexes[VRB_INDEX].get(
                productBaseName);
        return rc;
    }

    /**
     * Add a VirtualReductionBlock to the index.
     */
    private void addAssociatedVRB(
        String productBaseName,
        VirtualReductionBlockOBComponent vrb) {
        indexes[VRB_INDEX].put(productBaseName, vrb);
    }

    /**
     * Remove a VirtualReductionBlock from the index.
     */
    public void removeAssociatedVRB(String productBaseName) {
        indexes[VRB_INDEX].remove(productBaseName);
    }

    /**
     * Utility method.
     * @param from
     * @param to
     */
    private static void copyVector(Vector from, Vector to) {
        if (from == null || from.isEmpty() || to == null) {
            return;
        }
        Enumeration e = from.elements();
        while (e.hasMoreElements()) {
            Object obj = e.nextElement();
            if (obj != null && !to.contains(obj)) {
                to.addElement(obj);
            }
        }
    }

    /**
     * Utility/report method
     * @param ht
     */
    private void displayHashKeys(SortedHashtable ht) {
        Enumeration enum = ht.elements();
        // System.out.print("(PROG "+ht.toString());
        System.out.println("( ");
        while (enum.hasMoreElements()) {
            String tmpString = (String) enum.nextElement().toString();
            System.out.print(tmpString + " ");
        }
        System.out.println(")");
    }

    /**
     * true if the data-set is empty
     */
    public boolean isModelEmpty() {
        boolean rc = false;
        Vector tmpVector = getFileSet();
        if (tmpVector.size() == 0
            || (((Vector) tmpVector.elementAt(0)).size() == 0
                && tmpVector.size() == 1)) {
            rc = true;
        }
        return rc;
    }

    public void provideFrameInfo(
        KeywordContainerOBComponent hhoc,
        FrameInfo fi) {
        //String outObId, String outProgId, 
        //String outTelescopeId ) { 
        if (hhoc.hasIntermediate()) {
            InterContainer inter =
                (InterContainer) hhoc.getObservation(GasProp.TREE_DEFAULT);
            fi.obsId = inter.getObservation().getId();
            fi.programId = inter.getObservation().getProgram().getId();
            inter =
                (InterContainer) hhoc.getObservation(GasProp.TREE_TELESCOPE);
            fi.telescopeId =
                inter.getObservation().getProgram().getTelescope().toString();
        } else {
            Observation tmpObs = hhoc.getObservation(GasProp.TREE_DEFAULT);
            fi.obsId = tmpObs.getId();
            fi.programId = tmpObs.getProgram().getId();
            tmpObs = hhoc.getObservation(GasProp.TREE_TELESCOPE);
            fi.telescopeId = tmpObs.getProgram().getTelescope().toString();
        }
        //System.out.println("Telescope "+outTelescopeId + " Program "+outProgId+ " Obs "+outObId);
    }

    /**
     * to be used by selectedFilesToCommand
     */
    public String getClassificationForName(String comp) {
        String rc = null;
        OBComponent obc = (OBComponent) getObComponentByFakeName(comp);
        if (obc instanceof FitsFileOBComponent) {
            FitsFileOBComponent ffoc = (FitsFileOBComponent) obc;
            if (ffoc != null) {
                rc = ffoc.getClassification();
            } else {
                rc = new String("*ERROR*");
            }
        } else {
            rc = new String("NO-HEADER");
        }
        return rc;
    }
   

    /**
     * Implementation of the AssociationHandler interface.
     * Nodes are processed based on the following:
     *     - the nodes are FileAssociationNode objects
     *     - the files can be located in the model.
     *     - the file located are FileFileOBComponents.
     * 
     * @see org.eso.gasgano.services.data.AssociationHandler#processNodes(org.eso.gasgano.services.data.AssociationNode[])
     */
    public void processNodes(AssociationNode[] nodes) throws ServiceException {
        if (nodes == null) {
            return;
        }
        for (int i = 0; i < nodes.length; i++) {
            // process each node if it is a FileAssociationNode.
            AssociationNode node = nodes[i];
            if (node instanceof FileAssociationNode) {
                FileAssociationNode fnode = (FileAssociationNode) node;
                File address = fnode.getAddress();
                OBComponent obc = getObComponentByName(address.getPath());
                if (obc != null && obc instanceof KeywordContainerOBComponent) {
                    // node was found in the model.
                    classifyFileAssociationNode(
                        fnode,
                        (KeywordContainerOBComponent) obc);
                }
            }
        }
    }

    /**
     * Fill out the classification, grouping, etc of the FileAssociationNode
     * passed using the FileFIleOBComponent.
     * @param fan the association block to classify, group, etc.
     * @param ffoc
     */
    
    //TEST THAT TFITS CAN BE DRAG/DROPPED
    private void classifyFileAssociationNode(
        FileAssociationNode node,
        KeywordContainerOBComponent ffoc) {
            // set the classification
        node.setClassification(new String[] { ffoc.getClassification()});
            // set the instrument
        node.setInstrumentName(ffoc.getInstrument());
            // set the grouping
        node.setGrouping(new String[] {ffoc.getIntermediateGrouping()});
            // set the category
        node.setCategory(ffoc.isCalibFrame() ? "CALIB" : "RAW");
    }
    
    /**
     * Implementation if the AssociationHandler interface.
     * 
     * @see org.eso.gasgano.services.data.AssociationHandler#addFiles(java.io.File[])
     */
    public void addFiles(File[] files) throws ServiceException {
        for (int i = 0; i < files.length; i++) {
            // add the files to the model.
            serviceFileModel.addFile(files[i]);           
        }
        // update the model to reflect change in contents.
        update();
    }
    
    /**
     * Implementation if the AssociationHandler interface.
     * 
     * @see org.eso.gasgano.services.data.AssociationHandler#removeFiles(java.io.File[])
     */
    public void removeFiles(File[] files) {
        for (int i = 0; i < files.length; i++) {
            // add the file to the model.
            serviceFileModel.removeFile(files[i]);
        }
        // update the model to refelct the change in the contents.
        update();
    }
    
}
