/**
 * @author N. Kornweibel
 * @version $Id: ReductionBlockOBComponent.java,v 1.1 2003/10/29 15:23:16 nkornwei Exp $
 * A class defining the OBComponent of a ReductionBlock.
 */
package org.eso.gasgano.datamodel.gui;

import java.awt.Color;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.SwingConstants;

import org.eso.gasgano.datamodel.DataModelChangeListener;
import org.eso.gasgano.datamodel.DataModelChangeNotifier;
import org.eso.gasgano.datamodel.DataModelEvent;
import org.eso.gasgano.datamodel.FrameInfo;
import org.eso.gasgano.datamodel.ReductionBlock;
import org.eso.gasgano.datamodel.ReductionBlockCommon;
import org.eso.gasgano.datamodel.filesystem.DFSDataModel;
import org.eso.gasgano.datamodel.filesystem.DFSFile;
import org.eso.gasgano.gui.GUIUtils;
import org.eso.gasgano.properties.PropertyDB;

public class ReductionBlockOBComponent
    extends OBComponent
    implements DataModelChangeListener, ReductionBlockCommon {

    private ReductionBlock myFile = null;
    private DataModelChangeNotifier eventSource = null;

    private static JLabel dataLabel =
        new JLabel(
            new ImageIcon(
                ReductionBlockOBComponent.class.getResource(GUIUtils.dataGif)),
            (int) JLabel.LEFT_ALIGNMENT);

    /**
     * @param address The file name/path of the text file.
     */
    public ReductionBlockOBComponent(String address) {
        myFile = new ReductionBlock(address);
        insertInDataModel();
    }

    /**
     * Adding a ReductionBlockOBComponent into a DataModel:
     * Search for an entry in the data model which has the
     * same base name as the product name. If located, add this into the
     * DataModel using the prog and ob info for the located
     * component, and add this as a DataModelChange listener to 
     * that Component, so that the procedure is repeated if the 
     * data source changes.
     * If no file is located, add under "Unknown" catagories and add this
     * as a DataModelChangeListener to the DataModel, 
     * to repeat the procedure if new files are added.
     *
     * Extended by MZ to determine the telescope 
     */
    private void insertInDataModel() {
        DFSDataModel dm = DFSDataModel.getDataModel();
        String progId = null;
        String obId = null;
        String telescopeId = null;
        //
        DataModelChangeNotifier notifier = null;
        OBComponent dataSource = null;
        Vector dataSourceSet = null;

        if (myFile.getStatus() == DFSFile.FILE_INVALID) {
            return;
        }

        if (!dm.passFilter(this)) {
            myFile.setStatus(DFSFile.FILE_FILTERED);
            return;
        }
        dataSourceSet =
            dm.getObComponentByBaseName(
                myFile.getProductBaseName(),
                myFile.getProductDirectory());
        if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
            Enumeration e = dataSourceSet.elements();
            while (dataSource == null && e.hasMoreElements()) {
                dataSource = (OBComponent) e.nextElement();
                if (dataSource != null
                    && dataSource instanceof FitsFileOBComponent == false) {
                    // we only want to find a FitsFile.
                    dataSource = null;
                }
            }
        }
        if (dataSource == null) {
            progId = DFSDataModel.UNKNOWN_PROGRAM_ID;
            obId = DFSDataModel.UNKNOWN_OBSERVATION_ID;
            telescopeId = DFSDataModel.UNKNOWN_TELESCOPE_ID;

            notifier = dm;
        } else {
            FrameInfo frameInfo = new FrameInfo();
            dm.provideFrameInfo((KeywordContainerOBComponent) dataSource, frameInfo);

            telescopeId = frameInfo.telescopeId;
            progId = frameInfo.programId;
            obId = frameInfo.obsId;

            notifier = dataSource;
        }
        if (storeProgram != null
            && storeProgram.equals(progId)
            && storeObservation != null
            && storeObservation.equals(obId)
            && storeTelescope != null
            && storeTelescope.equals(telescopeId)) {
            // we have not changed prog/ob...
        } else {
            // remove this from the model if in the model
            // and add this, if valid.
            if (eventSource != null) {
                storeProgram = null;
                storeObservation = null;
                removeFromDataModel();
            }
            if (getStatus() == DFSFile.FILE_VALID) {
                String dir = myFile.getPath().getParent();
                dm.addObComponent(progId, obId, dir, telescopeId, this);
                storeProgram = (storeProgram == null) ? progId : storeProgram;
                storeObservation =
                    (storeObservation == null) ? obId : storeObservation;
                storeTelescope =
                    (storeTelescope == null) ? telescopeId : storeTelescope;
                registerToNotifier(notifier);
            }
        }

        // only add file associations if we are in the DataModel.
        if (eventSource != null) {
            // Add this Component as an associated ReductionBlock for 
            // all components read from the ReductionBlock.

            // this RB is associated to this.
            addAssociatedReductionBlock(this);
            // 1. Product associations:
            if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
                Enumeration e = dataSourceSet.elements();
                while (e.hasMoreElements()) {
                    OBComponent obc = (OBComponent) e.nextElement();
                    if (obc != null && obc.isPipeProduct()) {
                        obc.addAssociatedReductionBlock(this);

                    }
                }
            }
            // 2. Raw (input file) associations:
            dataSourceSet = myFile.getInputFileNames();
            if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
                Enumeration e = dataSourceSet.elements();
                while (e.hasMoreElements()) {
                    OBComponent obc =
                        dm.getObComponentByArcname((String) e.nextElement());
                    if (obc != null) {
                        obc.addAssociatedReductionBlock(this);
                    }
                }
            }
            // 3. Master Calib associations.
            dataSourceSet = myFile.getMasterCalibFileNames();
            if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
                Enumeration e = dataSourceSet.elements();
                while (e.hasMoreElements()) {
                    OBComponent obc =
                        dm.getObComponentByName((String) e.nextElement());
                    if (obc != null) {
                        obc.addAssociatedReductionBlock(this);
                    }
                }
            }
        } // end of if statement.
    } // end of insertInDataModel().

    /**
     * Remove the file associations - only to be called if this is
     * in the DataModel, as file associations are not built when this
     * is not a part of the DataModel.
     */
    private void removeFileAssociations() {
        DFSDataModel dm = DFSDataModel.getDataModel();
        Vector dataSourceSet =
            dm.getObComponentByBaseName(
                myFile.getProductBaseName(),
                myFile.getProductDirectory());

        // 1. Product associations:
        if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
            Enumeration e = dataSourceSet.elements();
            while (e.hasMoreElements()) {
                OBComponent obc = (OBComponent) e.nextElement();
                if (obc != null && obc.isPipeProduct()) {
                    obc.removeAssociatedReductionBlock(this);
                }
            }
        }
        // 2. Raw (input file) associations:
        dataSourceSet = myFile.getInputFileNames();
        if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
            Enumeration e = dataSourceSet.elements();
            while (e.hasMoreElements()) {
                OBComponent obc =
                    dm.getObComponentByName((String) e.nextElement());
                if (obc != null) {
                    obc.removeAssociatedReductionBlock(this);
                }
            }
        }
        // 3. Master Calib associations.
        dataSourceSet = myFile.getMasterCalibFileNames();
        if (dataSourceSet != null && dataSourceSet.isEmpty() == false) {
            Enumeration e = dataSourceSet.elements();
            while (e.hasMoreElements()) {
                OBComponent obc =
                    dm.getObComponentByName((String) e.nextElement());
                if (obc != null) {
                    obc.removeAssociatedReductionBlock(this);
                }
            }
        }
    }

    /**
     * Product files are all files matching the base product name
     * in the identical directory as the product name an which are
     * pipeline products (isPipeproduct()).
     */
    public Vector getProductFileNames() {
        DFSDataModel dm = DFSDataModel.getDataModel();
        Vector productFileNames = new Vector();
        Vector productOBComponents =
            dm.getObComponentByBaseName(
                myFile.getProductBaseName(),
                myFile.getProductDirectory());
        if (productOBComponents != null) {
            Enumeration e = productOBComponents.elements();
            while (e.hasMoreElements()) {
                OBComponent obc = (OBComponent) e.nextElement();
                // only PipeProducts can be RB Products.
                if (obc != null && obc.isPipeProduct()) {
                    productFileNames.addElement(obc.getId());
                }
            }
        }
        return productFileNames;
    }

    public Vector getInputFileNames() {
        return myFile.getInputFileNames();
    }

    public Vector getMasterCalibFileNames() {
        return myFile.getMasterCalibFileNames();
    }

    public int getStatus() {
        return myFile.getStatus();
    }

    /**
     * @param n The DataModelChangeNotifier to which this is to be
     * registered. Register this as a listener to the DataChangeNotifier 
     * removing the old registration if required. This method assumes there 
     * is only one DataChangeNotifier regisitered to at one time.
     */
    private void registerToNotifier(DataModelChangeNotifier n) {
        if (eventSource != n) {
            if (eventSource != null) {
                eventSource.removeDataModelChangeListener(this);
            }
            if (n != null) {
                n.addDataModelChangeListener(this);
            }
            eventSource = n;
        }
    }

    /**
     * @return The status following the update().
     * Check the status of the file and return it.
     * if the status has changed, re-read the file and notify
     * the appropriate listeners if applicable.
     * @see org.eso.gasgano.DFSFile
     */
    public int update() {
        if (myFile.dataRefreshRequired()) {
            // first we must remove the file from the model.
            if (getStatus() == DFSFile.FILE_VALID) {
                removeFromDataModel();
            }
            myFile.readData();
            insertInDataModel();
        }
        return getStatus();
    }

    public DFSFile getSourceFile() {
        return myFile;
    }

    /**
     * @param e The event to be processed.
     * Either the Component we used to get prog/ob info has
     * changed or we had no such Component and were in the Unknown group, 
     * and new files have been added to the DataModel,
     * either way, we need to remove
     * this from the data model and re-add this.
     */
    public void dataModelChanged(DataModelEvent e) {
        if (e.getEvent() == DataModelEvent.FILE_DELETED
            || e.getEvent() == DataModelEvent.FILE_ADDED
            || e.getEvent() == DataModelEvent.FILE_UPDATED) {
            if (PropertyDB.getInstance().isDebug(PropertyDB.debugEvent)) {
                System.out.println(getId() + " received dataModelChangeEvent.");
            }
            insertInDataModel();
        }
    }

    /**
     * @return The JComponent which encapsulates the data of this file.
     */
    public JComponent getDisplayComponent() {
        JComponent rc = null;
        if (myFile.readFile() == DFSFile.FILE_NOT_FOUND) {

            if (!DFSDataModel.getDataModel().isUpdating()) {
                new Thread(new Runnable() {
                    public void run() {
                        DFSDataModel.getDataModel().update();
                    }
                }).start();
            }

            rc =
                new JLabel(
                    "The file has been slyly removed from the filesystem",
                    SwingConstants.CENTER);
            return rc;
        }

        rc = new DFSTextFileDisplay(toString(), myFile.getData());
        // free the memory (as soon as the Model is finished with it.
        myFile.freeData();
        return rc;
    }

    public JComponent getTreeCellComponent() {
        JLabel cell = dataLabel;
        cell.setText(new String(toString()));
        cell.setForeground(Color.gray);
        cell.setFont(getLabelFont());
        return cell;
    }

    /**
     * @return The id string which uniquely identifies this in the
     * DataModel.
     * NB: This is the original address with which the FitsFile was created,
     * not the resolved address returned from a File object.
     */
    public String getId() {
        return myFile.getAddress();
    }

    public void cleanup() {
        super.cleanup();
        // remove our notification registration.
        registerToNotifier(null);
        removeFileAssociations();
    }

} // end of class ReductionBlockOBComponent.java

// end of file ReductionBlockOBComponent.java
