/**
 * @author N. Kornweibel
 * @version 1.0
 */

package org.eso.gasgano.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.border.BevelBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.eso.gasgano.datamodel.ExprFilter;
import org.eso.gasgano.datamodel.OBComponentFilter;
import org.eso.gasgano.keyword.EqKeywordExpr;
import org.eso.gasgano.keyword.ExpressionTable;
import org.eso.gasgano.keyword.Keyword;
import org.eso.gasgano.keyword.KeywordExpression;

public class FilterBuilder
    extends JDialog
    implements ActionListener, ListSelectionListener {

    private ExpressionBuilder exprBuilder = null;
    private JScrollPane tableView = null;
    private JTable table = null;
    private JSplitPane mainPane = null;
    private JPanel lowerPanel = null;
    private JPanel exprPanel = null;

    // expression and index of expression current selected.
    private KeywordExpression selectedExpression;
    private int selectedIdx = -1;

    private KeywordExpression emptyExpr;

    private JButton selectExpr = null;
    private JButton deleteExpr = null;
    private JButton copyExpr = null;
    private JButton cancelExpr = null;

    private ExpressionTable exprTable = null;

    public static void main(String args[]) {
        String tablePath = null;
        if (args.length < 1) {
            System.out.println("usage: <path to rule table>");
            System.exit(0);
        } else {
            tablePath = args[0];
        }
        ExpressionTable rt = ExpressionTable.readFile(tablePath);
        if (rt == null) {
            System.out.println(
                "Failed to read/locate: "
                    + tablePath
                    + " starting with an empty table.");
            rt = new ExpressionTable(tablePath);
        }

        FilterBuilder s =
            new FilterBuilder(null, "Filter Builder :" + tablePath, true, rt);

        s.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                System.exit(0);
            }
        });
        s.pack();
        // required by my window manager at least...
        s.setLocation(100, 300);
        s.show();
        System.out.println("Selection: " + s.getSelectedExpression());
        System.exit(0);
    }

    public KeywordExpression getSelectedExpression() {
        return selectedExpression;
    }

    public OBComponentFilter getSelectedFilter() {
        OBComponentFilter rc = null;
        if (selectedExpression
            != null // permit the empty expression 
        // && !selectedExpression.toString().equals(emptyExpr.toString())
        ) {
            rc = new ExprFilter(getSelectedExpression());
        }
        return rc;
    }

    public FilterBuilder(
        Frame parent,
        String title,
        boolean modal,
        ExpressionTable et) {
        super(parent, title, modal);
        exprTable = et;
        init(parent);
    }

    private KeywordExpression getEmptyExpression() {
        return new KeywordExpression(
            new EqKeywordExpr(
                new Keyword("KEYWORD", null),
                new String("value")));
    }

    private void init(Frame parent) {
        emptyExpr = getEmptyExpression();
        exprPanel = new JPanel(new BorderLayout());
        exprPanel.setBorder(new BevelBorder(BevelBorder.RAISED));

        table = new JTable(exprTable);
        table.getColumnModel().getColumn(0).setMinWidth(350);
        table.getColumnModel().getColumn(0).setPreferredWidth(350);

        ListSelectionModel lsm = table.getSelectionModel();
        if (lsm != null) {
            lsm.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
            lsm.addListSelectionListener(this);
        }

        tableView = new JScrollPane(table);
        tableView.setPreferredSize(new Dimension(600, 150));

        exprBuilder = new ExpressionBuilder(parent);
        exprBuilder.setPreferredSize(new Dimension(800, 150));
        exprBuilder.setBorder(new BevelBorder(BevelBorder.RAISED));

        selectExpr = new JButton("OK");
        selectExpr.addActionListener(this);
        copyExpr = new JButton("Duplicate");
        copyExpr.addActionListener(this);
        deleteExpr = new JButton("Delete");
        deleteExpr.addActionListener(this);
        cancelExpr = new JButton("Cancel");
        cancelExpr.addActionListener(this);
        lowerPanel = new JPanel();
        lowerPanel.add(selectExpr);
        lowerPanel.add(copyExpr);
        lowerPanel.add(deleteExpr);
        lowerPanel.add(cancelExpr);

        getContentPane().add(tableView, BorderLayout.NORTH);
        getContentPane().add(exprBuilder, BorderLayout.CENTER);
        getContentPane().add(lowerPanel, BorderLayout.SOUTH);
        updateButtons(false);
    }

    /**
     * List selection call back.
     * @see javax.swing.event.ListSelectionListener#valueChanged(ListSelectionEvent)
     */
    public void valueChanged(ListSelectionEvent e) {
        int i = -1;
        if (e != null && e.getValueIsAdjusting() == false) {
            // get the new selection by the user
            int newSelectedIdx;
            ListSelectionModel lsm = (ListSelectionModel) e.getSource();
            if (lsm.isSelectionEmpty()) {
                //no rows are selected
                newSelectedIdx = -1;
            } else {
                newSelectedIdx = lsm.getMinSelectionIndex();
            }
            commitEdits();

            if (newSelectedIdx < 0) {
                //no rows are selected
                selectedExpression = null;
                selectedIdx = -1;
            } else {
                selectedExpression =
                    (KeywordExpression) exprTable.elementAt(newSelectedIdx);
                selectedIdx = newSelectedIdx;
            }

            updateButtons(selectedExpression != null);
            displayExpression(selectedExpression);
        }
    }

    /**
     * Set selection sensitive buttons to enabled or disabled as required.
     * The OK button will always be enabled if the table is empty.
     * @param enabled
     */
    private void updateButtons(boolean enabled) {
        selectExpr.setEnabled(enabled);
        copyExpr.setEnabled(enabled);
        deleteExpr.setEnabled(enabled);
        selectExpr.setEnabled(enabled || exprTable.getRowCount() == 0);
    }

    /**
     * This method compares the contents of the expression builder with the 
     * expression table and if there are changes, saves the edited expression.
     * It should be called before closing the filter builder or loading a new
     * expression into the expression builder.
     * Does nothing if there are no changes.
     * Note that because changing the table model fires a table change event
     * that clears the current selection, this method explicitly sets the table
     * selection at the end of the method so that the current selection is displayed
     * in the GUI.
     */
    private void commitEdits() {
        KeywordExpression expr = exprBuilder.getKeywordExpression();
        if (selectedExpression != null) {
            if (expr != null) {
                String exprStr = expr.toString();
                boolean write = false;
                String selectedExprStr = selectedExpression.toString();
                write = !exprStr.equals(selectedExprStr);
                if (write) {
                    int selIdx = selectedIdx;
                    exprTable.replaceElement(selectedExpression, expr);
                    selectedIdx = selIdx;
                    selectedExpression =
                        (KeywordExpression) exprTable.elementAt(selectedIdx);
                }
            }
        }
    }

    /**
     * Write the expression/filter file, displaying a message to the user
     * on error.
     */
    private void writeFile() {
        try {
            exprTable.writeFile();
        } catch (IOException e) {
            userMessage("Error saving file: " + exprTable.getTablePath());
        }
    }

    /**
     * Button press call back.
     * implementation of the ActionListener interface
     */
    public void actionPerformed(ActionEvent a) {
        if (a.getSource() == selectExpr) {
            if (exprTable.getRowCount() != 0) {
                commitEdits();
            } else {
                selectedExpression = exprBuilder.getKeywordExpression();
                exprTable.addElement(selectedExpression);
                selectedIdx = 0;
            }
            writeFile();
            KeywordExpression expr = exprBuilder.getKeywordExpression();
            if (expr != null) {
                selectedExpression = exprTable.elementAt(selectedIdx);
                setSelectedRow(selectedIdx);
                displayExpression(selectedExpression);
                setVisible(false);
                dispose();
            } else {
                userMessage("Error, no expression found in builder.");
            }
        } else if (a.getSource() == deleteExpr) {
            if (selectedExpression != null) {
                // make a copy of the selectedIdx as it will be
                // set to -1 when we remove the element.
                int idx = selectedIdx;
                exprTable.removeElement(selectedExpression);
                writeFile();
                // re-set the selectedIdx.
                selectedIdx = idx;
                if (selectedIdx >= exprTable.getRowCount()) {
                    selectedIdx = exprTable.getRowCount() - 1;
                }
                if (selectedIdx >= 0) {
                    selectedExpression = exprTable.elementAt(selectedIdx);
                } else {
                    selectedExpression = null;
                    selectedIdx = -1;
                }
                setSelectedRow(selectedIdx);
                displayExpression(selectedExpression);

            } else {
                userMessage("No expression selected to delete");
            }
        } else if (a.getSource() == copyExpr) {
            if (selectedExpression != null) {
                selectedExpression = selectedExpression.copy();
                exprTable.addElement(selectedExpression);
                writeFile();
                selectedIdx = exprTable.getRowCount() - 1;
                setSelectedRow(selectedIdx);
                displayExpression(selectedExpression);
            } else {
                userMessage("No expression selected to copy");
            }
        } else if (a.getSource() == cancelExpr) {
            selectedExpression = null;
            selectedIdx = -1;
            table.clearSelection();
            // user wants to cancel, hide the window and dispose.
            setVisible(false);
            dispose();
        }
    }

    /**
     * Select the specified row in the table and ensure it is visible in the
     * scroll pane.
     * Does nothing if the row is out of range.
     * @param row
     */
    private void setSelectedRow(final int row) {
        if (row >= 0 && row < table.getRowCount()) {
            SwingUtilities.invokeLater(new Runnable() {
                public void run() {
                    table.getSelectionModel().setSelectionInterval(row, row);
                    table.scrollRectToVisible(table.getCellRect(row, 0, false));
                }
            });
        }
    }
    private void displayExpression(KeywordExpression c) {
        if (c == null || c.getExpressionTree() == null) {
            exprBuilder.setKeywordExpression(emptyExpr);
        } else {
            exprBuilder.setKeywordExpression(c);
        }
    }

    public void userMessage(String msg) {
        JOptionPane.showMessageDialog(this, msg);
    }

    /**
     * @see java.awt.Component#show()
     */
    public void setVisible(boolean vis) {
        if (vis && selectedIdx < 0 && exprTable.getRowCount() > 0) {
            selectedIdx = 0;
            setSelectedRow(selectedIdx);
            displayExpression(selectedExpression);
        }
        super.setVisible(vis);
    }

}

// end of file FilterBuilder.java
