/**
 * @author N. Kornweibel
 * @version 1.0
 * ExpressionTable - collection of keyword expressions.
 * TODO: this class requires swing as it defines the base table model.
 * would be good to use a table model that was not connected to 
 * GUI components.
 */

package org.eso.gasgano.keyword;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;

import org.eso.dfs.io.UpgradingObjectInputStream;
import org.eso.gasgano.properties.PropertyDB;

public class ExpressionTable
    extends AbstractTableModel
    implements Serializable {
        
    static final long serialVersionUID = -3380960670794527145L;
    
    private Vector expressions = new Vector();
    private static String columnNames[] = { "Expression" };
    private String tablePath = null;

    /**
     * Build an (empty) expression table using the given file (containing serialized
     * keyword expressions). The file is not loaded until readFile() is called.
     * @param filePath
     */
    public ExpressionTable(String filePath) {
        tablePath = filePath;
    }

    /**
     * Part of the AbstractTableModel extension.
     * @param c The column number.
     * @return The name of the column requirested.
     */
    public String getColumnName(int c) {
        return columnNames[c];
    }

    /**
     * Get the file associated with this expression table (where it will be saved to
     * and loaded from).
     * @return String the file path associated with this expression table.
     */
    public String getTablePath() {
        return tablePath;
    }

    /**
     * @return The number of columns in the table - always 1.
     */
    public int getColumnCount() {
        return 1;
    }

    /**
     * @return The number of rows (rules) in the table.
     */
    public int getRowCount() {
        return expressions.size();
    }

    /**
     * @param r The row number of the cell value required.
     * @param c The column number of the cell value required.
     * @return The Object represnting the value of the cell.
     */
    public Object getValueAt(int r, int c) {
        KeywordExpression expr = (KeywordExpression) expressions.elementAt(r);
        String rc = null;
        if (expr != null) {
            rc = expr.toString();
        }
        return rc;
    }

    /**
     * @param column The column whose width is required.
     * @return The width of the requested column.
     */
    public int getColumnWidth(int column) {
        return 30;
    }

    /**
     * @return False - the table is never editable.
     */
    public boolean isCellEditable(int r, int c) {
        return false;
    }

    /**
     * @return An Enumeration of the rules in the table.
     */
    public Enumeration elements() {
        return expressions.elements();
    }

    /**
     * @param o The rule to be removed from the table.
     * @return The true if the rule was located and removed.
     */
    public final synchronized boolean removeElement(Object o) {
        int idx = expressions.indexOf(o);
        if (idx >= 0) {
            expressions.removeElementAt(idx);
            fireTableRowsDeleted(idx, idx);
        }
        return idx >= 0;
    }

    /**
     * @param index The number of the rule required.
     * @return The KeywordExpression at the requested element, or null if
     * the element is empty/out of range.
     */
    public KeywordExpression elementAt(int index) {
        KeywordExpression rc = null;
        try {
            rc = (KeywordExpression) expressions.elementAt(index);
        } catch (ArrayIndexOutOfBoundsException e) {
            System.out.println(e.toString());
        }
        return rc;
    }

    /**
     * @param index The number of the rule to be removed.
     */
    public final synchronized void removeElementAt(int index) {
        try {
            expressions.removeElementAt(index);
        } catch (ArrayIndexOutOfBoundsException e) {
            System.out.println(e.toString());
        }
        fireTableRowsDeleted(index, index);
    }
    
    /**
     * Replace the old given object with the new given object
     * @param oldObj The rule to be removed from the table.
     * @param newObj the new rule to be added/inserted in the table
     * @return The true if the rule was located and replaced.
     */
    public final synchronized boolean replaceElement(Object oldObj, Object newObj) {
        int idx = expressions.indexOf(oldObj);
        boolean rc;
        if (idx >= 0) {
            expressions.set(idx, newObj);
            rc = true;
            fireTableRowsUpdated(idx, idx);
        } else {
            rc = false;
        }
        return rc;
    }

    /**
     * @param obj The rule to be added.
     */
    public final synchronized void addElement(Object obj) {
        expressions.addElement(obj);
        fireTableRowsInserted(expressions.size()-1, expressions.size()-1);
    }

    /**
     * @param obj The rule to be added.
     * @param index The index where the rule is to be stored.
     */
    public final synchronized void setElementAt(Object obj, int index) {
        try {
            expressions.setElementAt(obj, index);
        } catch (ArrayIndexOutOfBoundsException e) {
            System.out.println(e.toString());
        }
        fireTableRowsUpdated(index, index);
    }

    /**
     * Read a ExpressionTable in from a hard coded file name storing 
     * the Object stream.
     * @return The ExpressionTable read, or null if the read fails.
     */
    public static ExpressionTable readFile(String path) {
        ExpressionTable rc = new ExpressionTable(path);
        try {
            ObjectInputStream in =
                new UpgradingObjectInputStream(new FileInputStream(path),
            "org.eso.gasgano.", "org.eso.gasgano.keyword.");
            rc.expressions = (Vector) in.readObject();
        } catch (ClassNotFoundException e) {
            if (PropertyDB.getInstance().isDebug(PropertyDB.debugExpressionAndRules)) {
                System.out.println(
                    "Error reading file: " + path + " " + e.toString());
            }
            rc = null;
        } catch (IOException e) {
            if (PropertyDB.getInstance().isDebug(PropertyDB.debugExpressionAndRules)) {
                System.out.println(
                    "Error reading file: " + path + " " + e.toString());
            }
            rc = null;
        }
        return rc;
    }

    /**
     * Write a ExpressionTable out to a hard coded file name storing the Object
     * stream.
     */
    public void writeFile() throws IOException {
        ObjectOutputStream out =
            new ObjectOutputStream(new FileOutputStream(getTablePath()));
        out.writeObject(expressions);
        out.close();
    }
}
