package org.eso.gasgano.properties;

/**
 * Table for storing instrument specific keyword lists used for
 * grouping product.
 * 
 * The first column ("instrument") contains instrument names.
 * 
 * The second column ("criteria") contains a list of Keywords, seperated either by
 * commas or semicolons.
 * 
 * When save to the model (through setValue()) the keyword names are
 * checked for syntactic validity, and the semicolon seperators are replaced
 * by commas.
 */


import org.eso.gasgano.keyword.Keyword;

public class InstGroupTableModel extends PropertyTableModel {

    private GasProp gpx = null;
    private String[] menuLabels = new String[4];
    private String[] menuCommands = new String[4];
    
    private Boolean[] numbering = null;

    public InstGroupTableModel(GasProp gp) {
        loadProperties(gp);
    }

    public void loadProperties(GasProp gp) {
        gpx = gp;
        if (gpx.instruments.size() == 0) {
            insertRow(0);
        }
        numbering = new Boolean[gpx.instruments.size()];

        for (int i = 0; i < gp.instruments.size(); i++) {
            numbering[i] = new Boolean(false);
        }

        fireTableDataChanged();
    }

    public int getColumnCount() {
        return 2;
    }

    public int getRowCount() {
        return gpx.instruments.size();
    }

    public int getColumnWidth(int column) {
        if (column == 0 || column == 2) {
            return 20;
        }
        return 50;
    }

    public Object getValueAt(int row, int col) {
        Object rc = null;
        if (col == 0) {
            rc = gpx.instruments.elementAt(row);
        }
        if (col == 1) {
            rc = gpx.instrumentGrouping.elementAt(row);
        }
        if (col == 2) {
            rc = numbering[row];
        }
        return rc;
    }

    public boolean isCellEditable(int row, int col) {
        if (col == 2 && gpx.instrumentGrouping.elementAt(row).equals("")) {
            return false;
        }
        return true;
    }

    public String getColumnName(int column) {
        String rc = null;
        if (column == 0) {
            rc = "instrument";
        }
        if (column == 1) {
            rc = "keyword(s) defining group";
        }
        if (column == 2) {
            rc = "auto-num.";
        }
        return rc;
    }

    /**
     * Value is expected to be either an instrument (col 0) or a
     * list of keywords, seperated by commas or semi-colons.
     * @see javax.swing.table.TableModel#setValueAt(Object, int, int)
     */
    public void setValueAt(Object aValue, int row, int column) {
        switch (column) {
            case 0 :
                gpx.instruments.setElementAt(aValue, row);
                break;

            case 1 :
                // 1. replace all semicolons with commas
                // 2. split string into substring keyword names
                // 3. check keyword syntax
                // 4. set to properties if all is correct.
                String val = (String)aValue;
                val = val.replace(';', ',');
                String kwords[] = val.split(",");
                boolean valid = true;
                for (int i = 0; valid && i < kwords.length; i++) {
                    String kw = kwords[i].trim();
                    if (kw.length() > 0) {
                        valid = Keyword.validName(kw);
                    }
                }
                if (valid) {
                    gpx.instrumentGrouping.setElementAt(val, row);
                }
                break;

            case 2 :
                numbering[row] = (Boolean) aValue;
                break;
        }
    }

    public Class getColumnClass(int col) {
        Class cl = null;
        switch (col) {
            case 0 :
                cl = String.class;
                break;
            case 1 :
                cl = String.class;
                break;
            case 2 :
                cl = Boolean.class;
                break;
        }
        return cl;
    }

    public void removeRow(int rowNum) {
        try {
            gpx.instruments.removeElementAt(rowNum);
            gpx.instrumentGrouping.removeElementAt(rowNum);
            fireTableStructureChanged();
        } catch (ArrayIndexOutOfBoundsException e) {
            if (isDebug()) {
                System.out.println("GasPropEditor: array out of bound");
            }
        }
    }

    /**
     * 
     * @see org.eso.gasgano.properties.PropertyTableModel#insertRow(int)
     */
    public void insertRow(int rowNum) {
        try {
            gpx.instruments.insertElementAt("", rowNum);
            gpx.instrumentGrouping.insertElementAt("", rowNum);
            fireTableStructureChanged();
        } catch (ArrayIndexOutOfBoundsException e) {
            if (isDebug()) {
                System.out.println("GasPropEditor: array out of bound");
            }
        }
    }

    /**
     * Returns the numbering.
     * @return Boolean[]
     */
    public Boolean[] getNumbering() {
        return numbering;
    }

    /**
     * Sets the numbering.
     * @param numbering The numbering to set
     */
    public void setNumbering(Boolean[] numbering) {
        this.numbering = numbering;
    }

    /**
     * Not a test harness - just a small test program for the String.split() method.
     * @param args
     */
    public static void main(String[] args) {
        {
        String t1 = "1,2,3,456,7";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        } 
        {
        String t1 = "1, 2, 3  , 456 ,7";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        }
        {
        String t1 = "1,,t2,  ,t3,t456,7,";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        }
        {
        String t1 = "1";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        }
        {
        String t1 = ",";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        }
        {
        String t1 = "";System.out.println("\n<<<"+t1+">>>\n");
        String[] split = t1.split(",");
        for (int i = 0; i < split.length; i++) {
            System.out.println("<"+split[i]+">");
        }
        }
    }
}