package org.eso.gasgano.viewers;

/**
 * @author nkornwei
 *
 * This Singleton Factory is used to create
 * instances of ExternalViewer objects according
 * to the viewer types that are registered with
 * the factory.
 * 
 * The viewer types and classes are defined in
 * the ExternalViewers.properties resource bundle.
 */

import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.Vector;

public class ExternalViewerFactory {
    
    
    // the factory instance
    private static ExternalViewerFactory instance;
    
     /**
     * External Viewers bundle
     */
     private ResourceBundle viewers = ResourceBundle.getBundle("org.eso.gasgano.viewers.ExternalViewers");
    
    /**
     * Primary inteface to class - returns
     * the current instance of the factory.
     */
    public static synchronized ExternalViewerFactory getInstance() {
        if (instance == null) {
                instance = new ExternalViewerFactory();
        }
        return instance;
    }
    
    /**
     * Set the instance of the factory.
     */
    public static void setInstance(ExternalViewerFactory newInstance) {
        instance = newInstance;
    }     
    
    /**
     * Constructor 
     * - populates the viewer catagories and types.
     * - is private as this singleton is accessed through
     * - the gatInstance() method.
     */
    private ExternalViewerFactory() {
    }

    /**
     * Create and return an instance of the
     * an ExternalViewer based on the viewer
     * type specified.
     */
    public ExternalViewer getExternalViewer(String viewerType) throws ExternalViewerException {
        String viewerClass = viewers.getString(viewerType);
        ExternalViewer viewer = null;
        try {
            viewer = (ExternalViewer)Class.forName(viewerClass).newInstance();
        } catch(ClassNotFoundException cnfe) {
            // error in bundle
            throw new ExternalViewerException("Unknown external viewer: " + viewerClass);
        } catch(InstantiationException ie) {
            // error in bundle
            throw new ExternalViewerException("Error instantiating external viewer: " + viewerClass + " (InstantiationException)");
        } catch(IllegalAccessException iae) {
            // error in bundle
            throw new ExternalViewerException("Error instantiating storage translator: " + viewerClass + " (IllegalAccessException)");
        }
        return viewer;
    }
    
    /**
     * @return a Vector of all supported viewer types.
     */
    public Vector getViewerTypes() {
        Vector rc = new Vector();
        for (Enumeration e = viewers.getKeys(); e.hasMoreElements(); ) {
            rc.addElement((String)e.nextElement());
        }
        return rc;
    }
}
