package org.eso.gasgano.viewers;

/**
 * @author nkornwei
 *
 * Interface (via TCP sockets) to the SkyCat external viewer.
 * 
 * The address of the SkyCat port can either be specified or read 
 * from the $HOME/.rtd-remote file, created by SkyCat when
 * launched.
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.StringTokenizer;

public class SkyCatConnection {

    private InetAddress address = null;
    private int port = -1;
    private int pid = -1;
    // time to wait (in seconds) for
    // SkyCat to launch.
    private int timeout = 20;

    public static final String RtdPortFile = ".rtd-remote";

    /**
     * Create a SkyCatConnection with no
     * specified address.
     */
    public SkyCatConnection() {
    }

    /**
     * Specify the connection to connect to.
     */
    public SkyCatConnection(InetAddress addr, int pt) {
        setAddress(addr, pt);
    }
    
    /**
     * Set the time (in seconds) that the
     * connection waits when launching SkyCat.
     */
    public void setStartUpTimeout(int s) {
        timeout = s;
    }
    
    /**
     * Get the time (in seconds) that
     * the connection waits when launching SkyCat.
     */
    public int getStartUpTimeout() {
        return timeout;
    }

    /**
     * Set the SkyCat port address.
     */
    public void setAddress(InetAddress addr, int pt) {
        address = addr;
        port = pt;
    }

    /**
     * @return the port number.
     */
    public int getPort() {
        return port;
    }
    
    /**
     * @return the pid of the SkyCat process
     * as read from the file
     */
    public int getPid() {
        return pid;
    }

    /**
     * @return the host address
     */
    public InetAddress getAddress() {
        return address;
    }
    
    /**
     * Start sky cat with the specified command.
     */
    public void startSkyCat(String startCommand) throws IOException {
        String userHome = System.getProperties().getProperty("user.home");
        if (userHome == null || userHome.length() == 0) {
            throw new IOException("Could not determine user home directory, required to launch SkyCat.");
        }
        File rtdPortFile = new File(userHome + File.separatorChar + RtdPortFile);
        long lastMod = rtdPortFile.lastModified();
        
        runCommand(startCommand, true);
        
        boolean running = false;
        int remainingSleep = timeout*1000;
        long shortSleep = 500;
        do {
            running = lastMod != rtdPortFile.lastModified();
            try {
                Thread.sleep(shortSleep);
                remainingSleep -= shortSleep;
            } catch (InterruptedException e) {
                System.out.println("Warning: Exception during thread sleep.");
                // do nothing here.
            }
            if (remainingSleep <= 0) {
                break;
            }
        } while (!running);
        if (!running) {
            throw new IOException("Timeout waiting for SkyCat to launch after waiting "+timeout+" seconds.");
        }
    }
    
    /**
     * @param command The command and arguments to be executed.
     * @return The return code of the completed execution, will be -1
     * if the process is launched in the backGround.
     */
    private static int runCommand(String command, boolean backGround) throws IOException {
        int rc = -1;
        Runtime r = Runtime.getRuntime();
        Process p = r.exec(command);
        try {
            if (!backGround) {
                p.waitFor();
                rc = p.exitValue();
            }
        } catch (InterruptedException ie) {
            System.out.println(
                "Interrupted while executing: " + ie.toString());
            throw new IOException(ie.getMessage());
        } catch (IllegalThreadStateException ie) {
            System.out.println(
                "Interrupted while executing: " + ie.toString());
            throw new IOException(ie.getMessage());
        }
        return rc;
    }

    /**
     * Read the address from the .rtd-remote file.
     * The format of the file is: 
     *     PID HOST PORT
     * For example:
     *     21417 dmdsol8a 39554
     * @throws IOException if the file was not found or could
     * not be correctly read.
     */
    public void setAddressFromRTDFile() throws IOException {
        String userHome = System.getProperties().getProperty("user.home");
        if (userHome == null || userHome.length() == 0) {
            throw new IOException("Could not determine user home directory");
        }
        setAddressFromRTDFile(userHome + File.separatorChar + RtdPortFile);
    }

    /**
     * Read the address from the specified rtd-remote file.
     */
    public void setAddressFromRTDFile(String filePath) throws IOException {
        File theFile = new File(filePath);

        FileReader dataSource = null;
        try {
            dataSource = new FileReader(filePath);
        } catch (FileNotFoundException ex) {
            throw new IOException("File not found: " + filePath);
        }
        BufferedReader br = new BufferedReader(dataSource);
        String s = br.readLine();
        dataSource.close();
        if (s == null) {
            throw new IOException("Error reading file: " + filePath);
        }
        StringTokenizer st = new StringTokenizer(s);
        if (st.countTokens() < 3) {
            throw new IOException("Unexpected data in file: " + filePath);
        }
        // first read the PID 
        String pidStr = st.nextToken();

        // second read the host name
        String host = st.nextToken();

        // third read the port number
        String portStr = st.nextToken();

        try {
            port = Integer.parseInt(portStr);
            pid = Integer.parseInt(pidStr);
        } catch (NumberFormatException e) {
            throw new IOException(
                "Unable to read port or pid number from file: " + filePath);
        }
        try {
            address = InetAddress.getByName(host);
        } catch (UnknownHostException e) {
            throw new IOException("Unknown host: " + host);
        }
    }

    /**
     * Connect to the specified port and address.
     */
    public Socket connect() throws IOException {
        return new Socket(address, port);
    }

    /**
     * Send command to SkyCat - open a socket, send the command
     * read the reply, close the socket and return the reply.
     */
    public String sendCommand(String command) throws IOException {
        command = command + "\n";
        StringBuffer rc = new StringBuffer();
        Socket s = connect();
        OutputStream out = s.getOutputStream();
        out.write(command.getBytes());
        out.flush();
        BufferedReader rdr =
            new BufferedReader(new InputStreamReader(s.getInputStream()));
        int c;
        do {
            c = rdr.read();
            //System.out.println("READ: "+c);
            // end of stream
            if (c == -1) {
                break;
            }
            // new line character - end of message - right?
            if (c == 10) {
                break;
            }
            rc.append((char) c);
        } while (true);
        rdr.close();
        out.close();
        s.close();
        System.out.println("SkyCat Request :" + command);
        System.out.println("SkyCat Response:" + rc.toString());
        return rc.toString();
    }

    /**
     * Check if SkyCat is running - determine if the process
     * is running and see if a connection to the socket can be made.
     * @return true if it is running
     * @throws IOException if there is an error determining its state.
     */
    public boolean isSkyCatRunning() throws IOException {
        boolean rc = false;
        if (pid == -1 || port == -1 || address == null) {
            throw new IOException("Insufficient data to determine if SkyCat is running.");
        }
        // ps will return 0 if the process exists.
        // assumes /bin/ps is available on all platforms.....
        if (runCommand("/bin/ps -p "+pid, false) == 0) {
            try {
                Socket conn = connect();
                conn.close();
                // process is running and we connected....
                rc = true;
            } catch (IOException e) {
                // failed to connect
                rc = false;
            }
        }
        
        return rc;
    }

    /**
     * Send a file to be displayed in SkyCat.
     * A full file path is required.
     */
    public void sendFile(String fullPathToFile) throws IOException {
        sendCommand("config -file " + fullPathToFile);
    }

    /**
     * Test this class
     */
    public static void main(String[] args) {
        SkyCatConnection c = new SkyCatConnection();
        try {
            c.setAddressFromRTDFile("//K:/" + RtdPortFile);
            System.out.println(
                "Result from command: "
                    + c.sendCommand(
                        "config -file /home/nkornwei/fits/2001-02-11/r.UVES.2001-02-12T06:32:26.522_0002.fits"));
        } catch (IOException e) {
            System.out.println(e.getMessage());
        }
        System.out.println(
            "Connection: " + c.getAddress() + " port: " + c.getPort());
    }
}
