package org.eso.ohs.core.dbb.server;

// $Id: DbbSqlJoinKeyStack.java,v 1.2 2004/04/08 12:09:51 tcanavan Exp $
import java.util.Stack;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.dbb.sql.DbbSqlTable;

/**
 * Class to handle stacks of DbbSqlKeys objects.
 * It extends the standard Stack class providing methods to find
 * tables and to find the shortest stack.
 *
 * @author Carlo Boarotto, ESO 11-Sep-1999
 * 
 */
public class DbbSqlJoinKeyStack extends Stack {

 
private static final String rcsid = "$Id: DbbSqlJoinKeyStack.java,v 1.2 2004/04/08 12:09:51 tcanavan Exp $";
private static Logger stdlog_ = 
    Logger.getLogger( DbbSqlJoinKeyStack.class );
 
    //----------------------------------------
    // Constructor
    //----------------------------------------

    /** Call the super class (Stack) constructor */
    public DbbSqlJoinKeyStack() {
	super();
    }


    //----------------------------------------
    // Public Methods
    //----------------------------------------

    /** 
     * Get a key stored with the given index.
     *
     * @param i	Index.
     *
     * @return DbbSqlKey object.
     *
     **/
    public DbbSqlKey getKeyAt(int i) {
	return (DbbSqlKey)(this.elementAt(i));
    }


    //----------------------------------------
    // Private methods
    //----------------------------------------

    /**
     * Get the shortest stack contained in the given stack vector.
     * This utility is used to build the shortest path between two
     * tables.
     *
     * @param stack_vector Vector of stacks
     *
     * @return the shortest stack. 
     **/
    private static DbbSqlJoinKeyStack getShortestStack(Vector stack_vector) {

	int min_len = 10000;
	int vect_len = stack_vector.size();
	DbbSqlJoinKeyStack shortest = new DbbSqlJoinKeyStack();

	for(int i=0; i<vect_len; i++) {


	    DbbSqlJoinKeyStack current = 
		(DbbSqlJoinKeyStack)(stack_vector.elementAt(i));

	   

	    int stack_len = current.size();

	    if(stack_len < min_len || shortest == null) {
		shortest = current;
		min_len = stack_len;
	    }
	}
   

	return shortest;
    }


    //----------------------------------------
    // Public methods
    //----------------------------------------

    /**
     * Check if the stack contains a key referring to the given table.
     *
     * @param table	Table name.
     *
     * @return Boolean value: true if contined, false otherwise.
     *
     **/
    public boolean containsTable(DbbSqlTable table) {

	int stack_len = this.size();
	boolean result = false;
    
	for(int k=0; k<stack_len; k++) {
      
	    DbbSqlKey key = getKeyAt(k);
      
	    if(table.equals(key.getPkTable()) ||
	       table.equals(key.getFkTable())) {
	
		result = true;
		break;
	    }
	}

	return result;
    }


    /**
     * Get a 2 elements array: the first element is a vector of all the
     * tables involved in the join; the second element is a vector
     * containing the sql statements building up the join clause.
     *
     * @return An array of 2 Vectors.  
     */
    public DbbSqlJoinInfo getJoinInfo() {

	Vector tables_vect = new Vector();
	Vector clauses_vect = new Vector();

	for(int i=0; i<this.size(); i++) {

	    DbbSqlKey key = getKeyAt(i);

	    DbbSqlTable key_table1 = key.getPkTable();
	    DbbSqlTable key_table2 = key.getFkTable();

	    if(!tables_vect.contains(key_table1))
		tables_vect.addElement(key_table1);

	    if(!tables_vect.contains(key_table2))
		tables_vect.addElement(key_table2);

	    /* Improvement: it gets two clauses, the direct one and its
	     * reverse in order to let the db optimizer to optimize them 
	     */
	    //clauses_vect.addElement(key.getDirectJoin());
	    //clauses_vect.addElement(key.getReverseJoin());
	    String[]  keyClauses = key.getJoinClauses();
	    for (int c = 0; keyClauses != null && c < keyClauses.length; c++) {
           clauses_vect.addElement(keyClauses[c]);
	    }
	}

	// Improuvement: should get two clauses, the direct one and its
	// reverse in order to optimize them
	DbbSqlJoinInfo ret = new DbbSqlJoinInfo(tables_vect,
						clauses_vect);
	return ret;
    }

  
    /**
     * Get the preferred stack contained in the given stack vector.
     * This utility is used to build the preferred path between two
     * tables.
     *
     * @param stack_vector Vector of stacks
     *
     * @return the preferred stack. 
     **/
    public static DbbSqlJoinKeyStack getPreferredStack(DbbSqlTable from,
						       DbbSqlTable to,
						       Vector stack_vector) {

	int vect_len = stack_vector.size();
	Vector new_stack_vector = new Vector();

	for(int i=0; i<vect_len; i++) {

	    DbbSqlJoinKeyStack current_stack = 
		(DbbSqlJoinKeyStack)(stack_vector.elementAt(i));


	    int key_stack_len = current_stack.size();

	    /* If a key is defined as preferred key add its
	     * DbbSqlJoinKeyStack to the new vector of stacks. 
	     */
	    for(int k=0; k<key_stack_len; k++) {

		DbbSqlKey current_key = current_stack.getKeyAt(k);

		if(current_key != null && 
		   current_key.isPreferredKey(from, to)) {

		    new_stack_vector.addElement(current_stack);

		    break;
		}
	    } 
	}
    
	if(new_stack_vector.size() != 0)
	    return getShortestStack(new_stack_vector);

	return getShortestStack(stack_vector);
    }
}
