package org.eso.ohs.dfs;

import gnu.regexp.RE;
import gnu.regexp.REException;
import gnu.regexp.REMatch;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.eso.ohs.core.utilities.Convert;
import org.eso.ohs.core.utilities.OHSConfig;
import org.eso.ohs.core.utilities.OHSKeys;
import org.eso.ohs.core.utilities.PafFile;
import org.eso.ohs.core.utilities.Paramfile;
import org.eso.ohs.instruments.ParamFileParameter;

/**
 * $Id: Target.java,v 1.62 2005/11/23 09:11:44 pnunes Exp $
 * 
 * <p>
 * <br>
 * @author Bob Kemp
 */

public class Target extends BusinessObject {

    /** Standard log channel */
    private static Logger stdlog_ = Logger.getLogger(Target.class);

    /** Max/Min milliseconds for RA -- corresponds to 23:59:59.999 */
    protected final static int MAX_MILLISECS_RA = 86400000;
    protected final static int MIN_MILLISECS_RA = 0;

    /** Max/Min milliseconds for DEC -- corresponds to 90:00:00.000 */
    protected final static int MAX_MILLISECS_DEC = 324000000;
    protected final static int MIN_MILLISECS_DEC = -324000000;

    /** Max/Min Proper Motion in RA */
    protected final static float MAX_PROP_RA = 99999.99F;
    protected final static float MIN_PROP_RA = -99999.99F;

    /** Max/Min Proper Motion in Dec */
    protected final static float MAX_PROP_DEC = 99999.99F;
    protected final static float MIN_PROP_DEC = -99999.99F;

    /** Max/Min Differential Velocity in RA */
    protected final static float MAX_DIFF_RA = 99999.99F;
    protected final static float MIN_DIFF_RA = -99999.99F;

    /** Max/Min Differential Velocity in Dec */
    protected final static float MAX_DIFF_DEC = 99999.99F;
    protected final static float MIN_DIFF_DEC = -99999.99F;

    /** Min size for Equinox */
    protected final static int MIN_EQUINOX = 4;

    /** Version id for object serialization */
    static final long serialVersionUID = -7526074509242805287L;

    /* Legal values for the objectClass field */
    /** Target is in this galaxy */
    public static String[] objectClassValues_ = { "Unknown" };

    /* Legal values for the objectClass field */
    /** Target is in this galaxy */
    public static String[] objectClassTooltips_ = {
    };

    /** The right ascension of the target, in milliarcseconds */
    private int ra_;

    /** The declination of the target, in milliarcseconds */
    private int dec_;

    /** Epoch to be used for this observation */
    private float epoch_ = 2000.0F;

    /** Reference to the instance of PafFile, which keeps the values of ParamFile
     */
    private transient PafFile paramPaf_ = null;

    /**
     * Reference system for the coordinates.
     * Stored as a letter followed by 4 digits.
     */
    private String equinox_ = "2000";

    /** Proper motion in right ascension of the target, in milliarcsecs */
    private float propRA_;

    /** Proper motion in the declination of the target, in milliarcsecs */
    private float propDec_;

    /**
     * Proper motion in right ascension, a floating point number
     * (arcsecs per hour).
     */
    private float diffRA_;

    /**
     * Differential motion in declination of the target, a floating
     * point number (arcsecs per hour).
     */
    private float diffDec_;

    /**
     * What kind of object are we observing: galactic or extragalactic?
     * See OBJCLASS_* constants defined above for legal values.
     */
    private String objectClass_;

    /**
     * TBD -- Magnitude of the target.
     */
    private float mag_; // TBD

    /**
     * TBD -- Color of the target.
     */
    private float color_; // TBD

    /**
     *  Information provided by the Target author to the observer
     */
    private String userComments_ = "";

    Target() {
    }

    /**
     * Set new value for ra_ property
     *
     * @param newValue the new value for the property
     */
    public void setRA(int newValue) {
        ra_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for ra_ property
     *
     * @return the current value of the property
     */
    public int getRA() {
        return ra_;
    }

    /**
     * Set new value for dec_ property
     *
     * @param newValue the new value for the property
     */
    public void setDeclination(int newValue) {
        dec_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for dec_ property
     *
     * @return the current value of the property
     */
    public int getDeclination() {
        return dec_;
    }

    /**
     * Set new value for epoch_ property
     *
     * @param newValue the new value for the property
     */
    public void setEpoch(float newValue) {
        epoch_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for epoch_ property
     *
     * @return the current value of the property
     */
    public float getEpoch() {
        return epoch_;
    }

    /**
     * Set new value for equinox_ property
     *
     * @param newValue the new value for the property
     */
    public void setEquinox(String newValue) {
        equinox_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for equinox_ property
     *
     * @return the current value of the property
     */
    public String getEquinox() {
        return equinox_;
    }

    /**
     * Set new value for propDec_ property. Rounds two digits after the period.
     *
     * @param newValue the new value for the property
     */
    public void setProperDeclination(float newValue) {
        propDec_ = (float)Convert.round(newValue,2); 
        	
        firePropertyChangeEvent();
    }

    /**
     * Get value for propDec_ property
     *
     * @return the current value of the property
     */
    public float getProperDeclination() {
        return propDec_;
    }

    /**
     * Set new value for propRA_ property. Rounds two digits after the period.
     *
     * @param newValue the new value for the property
     */
    public void setPropRA(float newValue) {
        propRA_ = (float)Convert.round(newValue,2); 
        firePropertyChangeEvent();

    }

    /**
     * Get value for propRA_ property
     *
     * @return the current value of the property
     */
    public float getPropRA() {
        return propRA_;
    }

    /**
     * Set new value for diffRA_ property. Rounds two digits after the period.
     *
     * @param newValue the new value for the property
     */
    public void setDiffRA(float newValue) {
        diffRA_ = (float)Convert.round(newValue,2); 
        firePropertyChangeEvent();
    }

    /**
     * Get value for diffRA_ property
     *
     * @return the current value of the property
     */
    public float getDiffRA() {
        return diffRA_;
    }

    /**
     * Set new value for diffDec_ property. Rounds two digits after the period.
     *
     * @param newValue the new value for the property
     */
    public void setDiffDec(float newValue) {
        diffDec_ = (float)Convert.round(newValue,2); 
        firePropertyChangeEvent();
    }

    /**
     * Get value for diffDec_ property
     *
     * @return the current value of the property
     */
    public float getDiffDec() {
        return diffDec_;
    }

    /**
     * Set new value for userComments_ property
     *
     * @param newValue the new value for the property
     */
    public void setUserComments(String newValue) {
        userComments_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for userComments_ property
     *
     * @return the current value of the property
     */
    public String getUserComments() {
        if (userComments_==null)
            return("");
        else
            return userComments_;
    }

    /**
     * Set new value for mag_ property
     *
     * @param newValue the new value for the property
     */
    public void setMag(float newValue) {
        mag_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for mag_ property
     *
     * @return the current value of the property
     */
    public float getMag() {
        return mag_;
    }

    /**
     * Set new value for color_ property
     *
     * @param newValue the new value for the property
     */
    public void setColor(float newValue) {
        color_ = newValue;
        firePropertyChangeEvent();
    }

    /**
     * Get value for color_ property
     *
     * @return the current value of the property
     */
    public float getColor() {
        return color_;
    }

    /**
     * Set the list of legal values for for the objectClassValues_
     * property.
     */
    public static void setObjectClassValues(String[] newValues) {
        objectClassValues_ = newValues;
    }

    /**
     * Get the list of legal values for for the objectClassValues_
     * property.
     *
     * @return Array of the legal values.  
     */
    public static String[] getObjectClassValues() {
        return objectClassValues_;
    }

    /**
     * Set the list of legal values for for the objectClassTooltips_
     * property.     
     */
    public static void setObjectClassTooltips(String[] newValues) {
        objectClassTooltips_ = newValues;
    }

    /**
     * Get the list of legal values for for the objectClassTooltips_
     * property.
     *
     * @return Array of the legal values.
     */
    public static String[] getObjectClassTooltips() {
        return objectClassTooltips_;
    }

    /**
     * Set a new value for objectClass_ property. Value is set only if
     * it matches one of the legal values for this property.
     *
     * @param newValue New value for the property
     */
    public void setObjectClass(String newValue) {

        String[] legal = getObjectClassValues();
        String valueToSet = legal[0];

        stdlog_.debug("setObjectClass: newValue: '" + newValue + "'");

        if (!(newValue == null)) {
            for (int j = 1; j < legal.length; j++) {

                // stdlog_.debug( "setObjectClass:   trying: '" + 
                //                    legal[j] + "'" );

                if (legal[j].trim().equals(newValue.trim())) {

                    // stdlog_.debug( "setObjectClass:    FOUND: '" + 
                    //                     legal[j] + "'" );

                    valueToSet = legal[j];
                    break;
                }
            }
        }
        objectClass_ = valueToSet;
        firePropertyChangeEvent();
    }

    /**
     * Get value for objectClass_ property
     *
     * @return the current value of the property
     */
    public String getObjectClass() {
        return objectClass_;
    }

    /**
     * Returns the string used to refer to this type of BusinessObject
     * in messages to the user.
     *
     * @return user-oriented name
     */
    public static String getDisplayName() {
        return "Target";
    }

    /**
     * Returns the filename suffix for files containing this type of object.
     *
     * @return suffix for filename when stored in local filesystem
     */
    public static String getSuffix() {
        return "tp";
    }

    /**
     * Override readObject to set any invalid values to within the
     * range in the 2.1.X cache
     *
     # <br>ChangeLog: A M Chavan, 10-Jul-2001: rewritten, added RA and Dec.
     */

    private void readObject(ObjectInputStream in)
        throws IOException, ClassNotFoundException {

        in.defaultReadObject();

        if (ra_ > MAX_MILLISECS_RA)
            ra_ = MAX_MILLISECS_RA;
        if (ra_ < MIN_MILLISECS_RA)
            ra_ = MIN_MILLISECS_RA;

        if (dec_ > MAX_MILLISECS_DEC)
            dec_ = MAX_MILLISECS_DEC;
        if (dec_ < MIN_MILLISECS_DEC)
            dec_ = MIN_MILLISECS_DEC;

        if (propRA_ > MAX_PROP_RA)
            propRA_ = MAX_PROP_RA;
        if (propRA_ < MIN_PROP_RA)
            propRA_ = MIN_PROP_RA;

        if (propDec_ > MAX_PROP_DEC)
            propDec_ = MAX_PROP_DEC;
        if (propDec_ < MIN_PROP_DEC)
            propDec_ = MIN_PROP_DEC;

        if (diffRA_ > MAX_DIFF_RA)
            diffRA_ = MAX_DIFF_RA;
        if (diffRA_ < MIN_DIFF_RA)
            diffRA_ = MIN_DIFF_RA;

        if (diffDec_ > MAX_DIFF_DEC)
            diffDec_ = MAX_DIFF_DEC;
        if (diffDec_ < MIN_DIFF_DEC)
            diffDec_ = MIN_DIFF_DEC;
    }

    /**
     * Verifies this Target, returning the concatenation of all error messages.
     * 
     * @return the (possibly empty) concatenation of all error messages.
     */
    public String verify() {
        RE eq_RE = null; // the regexp object
        REMatch eq_rem = null; // the regexp's result

        StringBuffer sb = new StringBuffer();
        /** cmavros Correct message (b or l) to (B or J) 
         *  [1464] Add equinox length check
         */
        final String equiErr =
            "The equinox must be at a minimum a\nfour digit whole number with the"
                + "optional leading letter B or J.\nAn optional decimal point "
                + "with a maximum of two digits is also allowed.\n";

        String eq_pattern = "(([BJ])?\\d{0,2}\\d{0,2}(\\.\\d{0,2})?)?";

        try {
            eq_RE = new RE(eq_pattern);
        } catch (REException e) {
            stdlog_.debug(
                "CLASS Target - verify(equinox): Autsch!!!!! "
                    + e.getMessage());
        }
        if (!eq_RE.isMatch(equinox_) || equinox_.length() < MIN_EQUINOX) {
            sb.append(equiErr);
        }

        /* AMC, 04-Jul-2001: fixed 1448 */
        if (!Convert.isValidRA(ra_))
            sb.append("The value of Right Ascension is out of range.\n");
        if (!Convert.isValidDec(dec_))
            sb.append("The value of Declination is out of range.\n");
        /* 1448 end */

        /* item 1372 begin */
        if (propRA_ > MAX_PROP_RA || propRA_ < MIN_PROP_RA) {
            sb.append(
                "The value of Proper Motion RA is out of range."
                    + "\nIt must be between "
                    + MIN_PROP_RA
                    + " and "
                    + MAX_PROP_RA
                    + ".\n");
        }
        if (propDec_ > MAX_PROP_DEC || propDec_ < MIN_PROP_DEC) {
            sb.append(
                "The value of Proper Motion DEC is out of range."
                    + "\nIt must be between "
                    + MAX_PROP_DEC
                    + " and "
                    + MIN_PROP_DEC
                    + ".\n");
        }
        if (diffRA_ > MAX_DIFF_RA || diffRA_ < MIN_DIFF_RA) {
            sb.append(
                "The value of Diff RA is out of range."
                    + "\nIt must  be between"
                    + MIN_DIFF_RA
                    + " and "
                    + MAX_DIFF_RA
                    + ".\n");
        }
        if (diffDec_ > MAX_DIFF_DEC || diffDec_ < MIN_DIFF_DEC) {
            sb.append(
                "The value of Diff DEC is out of range."
                    + "\nIt must  be between"
                    + MIN_DIFF_DEC
                    + "and "
                    + MAX_DIFF_DEC
                    + ".\n");
        }
        /* item 1372 end */
        return sb.toString();
    }


    public String verifyBO() {
        return verify();
    }


    /**
     * Generic hook for "visitor" Design Pattern.
     * See the BusinessVisitor interface for more details.
     * 
     * @param visitor	the class whose visit() method we wish to invoke.
     * 
     * @see BusinessVisitor
     */
    public Object accept(BusinessVisitor visitor, Object arg) {
        return visitor.visit(this, arg);
    }

    /**
     * Update this object with some values extracted from the input
     * <code>Paramfile</code>.
     * <P>
     * The keywords we look for are: <code></code>, TBD.
     *
     * @param pfp Where we look for meaningful keywords.
     *
     * @see #updateFromParamfile( Paramfile ) 
     *
     * <br>ChangeLog: A M Chavan 16-Aug-2001: creation
     * @exception IllegalArgumentException
     */

    public void updateFromParamFileParameter(ParamFileParameter pfp)
        throws IllegalArgumentException {

        if (!pfp.getValue().equals("NODEFAULT")) {
            Paramfile p = pfp.getParamfile();
            updateFromParamfile(p);
        }
    }

    /**
     * Update this object with some values extracted from the input
     * <code>Paramfile</code>.
     * <P>
     * The keywords we look for are are configured in site.cf (OHSConfig).
     * 
     * @param paramfile Where we look for meaningful keywords.
     *
     * @see #ra_
     * @see #dec_
     * @see #epoch_
     * @see #equinox_
     * @see #propRA_
     * @see #propDec_
     * @see #color_
     * @see #mag_
     * @see #name_
     *
     * <br>ChangeLog: A M Chavan 09-Aug-2001
     * <br>ChangeLog: A M Chavan 17-Aug-2001: added handling of 
     *                                 IllegalArgumentException
     *
     * @exception IllegalArgumentException
     */

    public void updateFromParamfile(Paramfile paramfile)
        throws IllegalArgumentException {

        IllegalArgumentException conversionException = null;
        // paramfile.report();
        suspendPropertyChangeEvents();

        String kval; // keyword value

        /*
         * TEL.TARG.ALPHA and TEL.TARG.RA are synonyms
         */
        /* left in for reference - remove later after 2.6 release
        kval = paramfile.getValue("TEL.TARG.ALPHA");
        if (kval.length() == 0) {
            kval = paramfile.getValue("TEL.TARG.RA");
        }
        kval = kval.trim();
        */
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetRAParam, paramfile)
                .trim();

        if (kval.length() > 0) {
            boolean conversionError = false;

            int ra_total = 0;
            ;
            try {
                ra_total = Convert.HHMMSSToMilliarcsec(kval);
            } catch (IllegalArgumentException e) {
                conversionError = true;
                conversionException = e; // remember this
            }

            if (!conversionError) {
                setRA(ra_total);
            }
        }

        /*
         * TEL.TARG.DELTA and TEL.TARG.DEC are synonyms
         */
        /* left in for reference - remove later after 2.6 release
        kval = paramfile.getValue("TEL.TARG.DELTA");
        if (kval.length() == 0) {
            kval = paramfile.getValue("TEL.TARG.DEC");
        }
        kval = kval.trim();
        */
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetDecParam, paramfile)
                .trim();
                
        if (kval.length() > 0) {
            boolean conversionError = false;

            int dec_total = 0;
            try {
                dec_total = Convert.DDMMSSToMilliarcsec(kval);
            } catch (IllegalArgumentException e) {
                conversionError = true;
                conversionException = e; // remember this
            }

            if (!conversionError) {
                setDeclination(dec_total);
            }
        }

        /*
         * TEL.TARG.EPOCH
         */
        // left in for reference - remove later after 2.6 release
        //kval = paramfile.getValue("TEL.TARG.EPOCH").trim();
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetEpochParam, paramfile)
                .trim();
                
        if (kval.length() > 0) {

            // Keep only the first 'maxEpochLenght' chars
            // of the epoch.
            // Note: the value of maxEpochLenght
            // comes from the DDD,
            // obrep2..target_packages.epoch
            //----------------------------------------
            int maxEpochLenght = 6;
            if (kval.length() > maxEpochLenght)
                kval = kval.substring(0, maxEpochLenght - 1);

            Float t = new Float(kval);
            float epoch = t.floatValue();
            setEpoch(epoch);
        }

        /*
         * TEL.TARG.EQUINOX
         */
        // left in for reference - remove later after 2.6 release
        //kval = paramfile.getValue("TEL.TARG.EQUINOX").trim();
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetEquinoxParam, paramfile)
                .trim();
        
        if (kval.length() > 0) {

            // Keep only the first 'maxEquinoxLenght' chars
            // of the equinox.
            // Note: the value of maxEquinoxLenght
            // comes from the DDD,
            // obrep2..target_packages.equinox
            //----------------------------------------
            int maxEquinoxLenght = 8;
            if (kval.length() > maxEquinoxLenght)
                kval = kval.substring(0, maxEquinoxLenght - 1);
            setEquinox(kval);
        }

        /*
         * TEL.TARG.PMA and TEL.TARG.PROPRA
         */
        /* left in for reference - remove later after 2.6 release
        kval = paramfile.getValue("TEL.TARG.PROPRA");
        if (kval.length() == 0)
            kval = paramfile.getValue("TEL.TARG.PMA");
        kval = kval.trim();
        */
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetPropRAParam, paramfile)
                .trim();


        if (kval.length() > 0) {
            Float t = new Float(kval);
            float pma = t.floatValue();
            setPropRA(pma);
        }

        /*
         * TEL.TARG.PMD and TEL.TARG.PRODEC
         */
        /* left in for reference - remove later after 2.6 release
        kval = paramfile.getValue("TEL.TARG.PRODEC");
        if (kval.length() == 0) {
            kval = paramfile.getValue("TEL.TARG.PMD");
        }
        kval = kval.trim();
        */
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetPropDecParam, paramfile)
                .trim();

        if (kval.length() > 0) {
            Float t = new Float(kval);
            float pmd = t.floatValue();
            setProperDeclination(pmd);
        }

        /*
         * TEL.TARG.COLOR
         */
        // left in for reference - remove later after 2.6 release
        //kval = paramfile.getValue("TEL.TARG.COLOR").trim();
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetColorParam, paramfile)
                .trim();
        if (kval.length() > 0) {
            Float t = new Float(kval);
            float color = t.floatValue();
            setColor(color);
        }

        /*
         * TEL.TARG.MAG
         */
        // left in for reference - remove later after 2.6 release
        //kval = paramfile.getValue("TEL.TARG.MAG").trim();
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetMagParam, paramfile)
                .trim();
        if (kval.length() > 0) {
            Float t = new Float(kval);
            float mag = t.floatValue();
            setMag(mag);
        }

        /*
         * TEL.TARG.NAME
         */
        // left in for reference - remove later after 2.6 release
        //kval = paramfile.getValue("TEL.TARG.NAME").trim();
        kval =
            OHSConfig
                .getInstance()
                .evalPropertyFromParamfile(OHSKeys.TargetNameParam, paramfile)
                .trim();

        if (kval.length() > 0) {
            setName(kval);
        }

        resumePropertyChangeEvents();
        firePropertyChangeEvent();

        if (conversionException != null)
            throw conversionException;
    }

    public static void initialiseBusinessObjects(File objClassFile) throws IOException {
    
        Vector classifications = new Vector();
        Vector tooltips = new Vector();
        BufferedReader rdr = null;
        String objClass = null, ttip = null;
        final int start_objClass = 0;
        final int end_objClass = 20;
        String line;
    
        try {
            rdr = new BufferedReader(new FileReader(objClassFile));
    
            for (;;) {
                line = rdr.readLine();
    
                if (line == null) {
                    break;
                }
    
                int index = line.indexOf(':');
    
                if (index < 0) {
                    continue;
                }
    
                line = line.substring(index + 1);
    
                if (line.trim().length() == 0) {
                    continue;
                }
    
                if (line.length() > end_objClass) {
                    int firstQuote = line.indexOf('"');
                    int lastQuote = line.lastIndexOf('"');
    
                    if (firstQuote < 0 || lastQuote < 0) {
                        continue;
                    }
    
                    objClass = line.substring(start_objClass, end_objClass);
                    ttip = line.substring(firstQuote + 1, lastQuote);
                } else {
                    continue;
                }
    
                classifications.addElement(objClass);
                tooltips.addElement(ttip);
            }
        } catch (IOException e) {
            //P2PP.stdlog_.error("", e);
        }
    
        if (rdr != null) {
            rdr.close();
        }
    
        String[] objClassValues = new String[classifications.size()];
        classifications.copyInto(objClassValues);
    
        Target.setObjectClassValues(objClassValues);
    
        String[] tooltipValues = new String[tooltips.size()];
        tooltips.copyInto(tooltipValues);
    
        Target.setObjectClassTooltips(tooltipValues);
    
    }
    /**
     * @see org.eso.ohs.dfs.StorableObject#postProecess()
     */
    public void postProecess() {
    }
    
    
    /**
     * Compares two <code>Target</code> objects, comparing some properties
     *
     * @param other
     * @return boolean
     */
    public boolean equals(Target other) {
        
        boolean res = true;
        
        if ( getDiffDec() != other.getDiffDec())
            res = false;
        else
            if ( getDiffRA() != other.getDiffRA())
                res = false;
            else
                if ( getEpoch() != other.getEpoch())
                    res = false;
                else
                    if ( getEquinox().compareTo(other.getEquinox()) != 0 )
                        res = false;
                    else
                        if ( getProperDeclination() != other.getProperDeclination())
                            res = false;
                        else
                            if ( getPropRA() != other.getPropRA())
                                res = false;
                            else
                                if ( getRA() != other.getRA())
                                    res = false;                    
                                else
                                    if ( getUserComments().compareTo(other.getUserComments()) != 0 )
                                        res = false;                    
              
        return (res);
        
    }
    
}
