/*
 * Created by amchavan on Aug 30, 2005
 */
package org.eso.ohs.phase2.timeline;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.Iterator;
import java.util.TreeSet;
import java.util.Vector;

import org.eso.ohs.phase2.visibility.PCF;
import org.eso.ohs.phase2.visibility.PCFInterval;

/**
 * Abstract superclass of all classes representing a
 * computed timeline.
 * 
 * Concrete subclasses must implement the {@link #compute()}
 * method. Constructors invoke the {@link #compute()} method directly,
 * so that right after construction one can invoke the 
 * {@link #getSchedTasks()} method to review the computed schedule.
 * 
 * @version $Revision: 1.6 $
 * @author amchavan, Aug 30, 2005
 */

// $Id: Timeline.java,v 1.6 2005/11/09 17:29:53 amchavan Exp $

public abstract class Timeline {

    /**
     * The value of the preference PCF indicating that observing in that
     * interval is not possible.
     */
    protected static final double NOGO = 2.5;

    /**
     * Lower priority numbers indicate higher priority
     * @see TimelineTask#compareTo(TimelineTask)
     */
    private class CompareTasksByPriority implements Comparator {
        public int compare( Object arg0, Object arg1 ) {
            TimelineTask t0 = (TimelineTask) arg0;
            TimelineTask t1 = (TimelineTask) arg1;
            return t0.compareTo( t1 );
        }
    }

    /** Lower priority numbers indicate higher priority */
    private class CompareTasksByDate implements Comparator {
        public int compare( Object arg0, Object arg1 ) {
            TimelineTask t0 = (TimelineTask) arg0;
            TimelineTask t1 = (TimelineTask) arg1;
            long t = t0.getStartTime().getTime() - t1.getStartTime().getTime();
            if( t < 0 )
                return -1;
            if( t > 0 )
                return 1;
            return 0;
        }
    }

    /** The list of Tasks we should schedule,
     *  sorted by task priority */
    protected TimelineTask[] tasks;
    
    /** The timeline is represented as a PCF whose "values" are task IDs */
    protected PCF timeline;
    
    /** The collection of all tasks we have already scheduled at any time */
    protected TreeSet schedTasks;
    
    /** 
     * This PCF represents the valid time range for this timeline; 
     * that is, the interval between the <em>start</em> and <em>end</em> 
     * parameters given to the constructor.
     * <br>
     * It looks like:
     * <pre>
     *       +-----------+
     * ------+           +---------
     * </pre>
     * where the PCF is 1.0 in the central interval, and 0.0 outside.
     */
    protected PCF range;

    /** Holds the <em>start</em> parameter given to the constructor. */
    protected Date start;

    /** Holds the <em>end</em> parameter given to the constructor. */
    protected Date end;
    
    /** Empty constructor (non-public, for testing only) */
    Timeline() {
        this( new TimelineTask[0], new Date(), new Date() );
    }
    
    /** 
     * Constructor
     * @param tasks  The list of tasks to put on the timeline, 
     *               can be empty but not <code>null</code>
     * @param start  The starting datetime for the timeline
     * @param end    The closing datetime for the timeline
     */
    public Timeline( TimelineTask[] tasks, Date start, Date end ) {

        assert start.getTime() != 0;
        
        this.timeline   = new PCF();
        this.schedTasks = new TreeSet( new CompareTasksByDate() );
        this.range      = new PCF();
        this.start      = start;
        this.end        = end;
        range.addInterval( start, end, 1.0 ); 
        
        // Count how many of the input tasks can be executed
        int doableCounter = 0;
        for( int i = 0; i < tasks.length; i++ ) 
            if( tasks[i].isDoable() )
                doableCounter++;
        
        // Filter out all non-executable tasks
        int j = 0;
        TimelineTask[] doables = new TimelineTask[ doableCounter ];
        for( int i = 0; i < tasks.length; i++ ) 
            if( tasks[i].isDoable() )
                doables[j++] = tasks[i];
        
        this.tasks = doables;
        Arrays.sort( this.tasks, new CompareTasksByPriority() );
        for( int i = 0; i < this.tasks.length; i++ ) 
            this.tasks[i].setStartTime( null );
        
        compute();
    }
    
    /** 
     * Constructor
     * @param tasks  The list of tasks to put on the timeline: all
     *               elements of the Vector should be instances of
     *               class TimelineTask. This parameter 
     *               can be empty but not <code>null</code>
     * @param start  The starting datetime for the timeline
     * @param end    The closing datetime for the timeline
     */
    public Timeline( Vector tasks, Date start, Date end ) {
        this( (TimelineTask[]) tasks.toArray( new TimelineTask[0] ), 
              start, 
              end );
    }
    
    /** Compute the timeline. */
    protected abstract void compute();
    
    // For testing only
    TimelineTask[] getTasks() {
        return tasks;
    }

    /** @return  All scheduled tasks: the array is sorted by start date */
    public TimelineTask[] getSchedTasks() {
        int n = getSchedTasksCount();
        TimelineTask[] ret = new TimelineTask[ n ];
        int i = 0;
        for( Iterator iter = schedTasks.iterator(); iter.hasNext(); ) {
            TimelineTask task = (TimelineTask) iter.next();
            ret[i++] = task;
        }
        return ret;
    }
    
    /** @return The number of tasks that were actually scheduled */
    public int getSchedTasksCount() {
        return schedTasks.size();
    }

    /** @return A packed PCF resulting from forcing the input task's
      *         preference PCF to <em>NOGO</em> 
      *         outside this timeline's range, as well as for 
      *         all intervals already taken up by previously
      *         scheduled tasks.
      *         
      * @param task  The task to process (remains unchanged)
      * @see #NOGO
      */
    protected PCF removeUnsuitableIntervals( TimelineTask task ) {
        
        // Border condition: can we accommodate this task at all?
        long taskMsec = task.getDuration() * 1000;        
        if( getRangeDuration() < taskMsec )
            return new PCF( NOGO );
        
        // Start from the raw preference and remove progressively from there
        PCF taskpref = (PCF) task.getPreference().clone();
        
        // Remove from our range this task's duration, then confine
        // the task's preference to that. 
        // Our timeline:          -----------================--------------
        // Task duration:                                ****
        // Possible /start/ time: -----------============------------------
        PCF reducedRange = reduceRange( taskMsec );
        taskpref.confine( reducedRange, NOGO );
        taskpref.pack();
        
        // Now we take the current timeline and extend every non-zero
        // interval backwards by the task's duration. In fact, our task
        // cannot *end* at any time that's already taken up.
        PCF extendedTimeline = extendSlotsBackwards( taskMsec );
        taskpref.exclude( extendedTimeline, NOGO );
        taskpref.pack();
        return taskpref;
    }

    /**
     * @return A PCF representing this timeline's range reduced
     *         by the input time; that is, the end time is shifted
     *         backwards.
     * @param time  Time in milliseconds
     */
    PCF reduceRange( long time ) {
        // Consistency check
        assert range.getIntervalCount() == 3;
        
        PCFInterval i = range.getIntervalAt( 1 );
        Date rangeStart = i.getStart();
        Date rangeEnd = i.getEnd();
        long oldend = rangeEnd.getTime();
        long newend = oldend - time;
        Date newRangeEnd = new Date( newend );
        PCF ret = new PCF( 0.0 );
        ret.addInterval( rangeStart, newRangeEnd, 1.0 );
        return ret;
    }

    /**
     * @return A packed PCF where every non-zero interval of this timeline
     *         is extended backwards by the input time; that is, every
     *         interval's start time is shifted backwards.
     *         
     * @param time  Time in milliseconds
     */
    PCF extendSlotsBackwards( long time ) {
        Vector intervals = new Vector();
        PCF extTimeline = (PCF) timeline.clone();
        int n = extTimeline.getIntervalCount();
        for( int i = 0; i < n; i++ ) {
            // NOTE: we loop over timeline but we update extTimeline
            PCFInterval interval = timeline.getIntervalAt( i );
            double value = interval.getValue();
            if( value == 0.0 )
                continue;
            
            Date intStart = interval.getStart();
            long oldStart = intStart.getTime();
            long newStart = oldStart - time;
            if( newStart < 0 )      // amchavan, 09-Nov-2005
                newStart = 0;
            Date newIntStart = new Date( newStart );
            boolean ok = extTimeline.setInterval( newIntStart, intStart, value );
            if( ok )
                continue;
            
            // we failed to add the interval because it overlaps two intervals
            // do something about it...
            assert ok;
        }
        extTimeline.pack();
        return extTimeline;
    }

    /**
     * Add the a task to this timeline, making the corresponding 
     * slot unsuitable for all future candidate tasks.
     * 
     * @param task   TimelineTask to add
     * @param start  Start date for the task
     */
    protected void addTask( TimelineTask task, Date start ) {

        // Precondition: task ID must be non-zero
        assert task.getID() != 0;
        
        task.setStartTime( start );
        Date end = new Date( start.getTime() + task.getDuration() * 1000 );
        boolean ok = timeline.addInterval( start, end, task.getID() );
        
        //===========================================================
        // TODO: debugging stuff, remove from production-quality code
        //       amchavan, 12-Oct-2005
        assert ok : "Internal error, should never happen";

//        System.out.println( "Timeline start: "  + this.start.getTime() + " " + this.start );
//        System.out.println( "Timeline end:   "  + this.end.getTime() + " " + this.end );

        Date tStart = task.getStartTime();
        Date tEnd   = task.getEndTime();
//        System.out.println( "tStart: " /* + tStart.getTime() + " " */ + tStart );
//        System.out.println( "tEnd: "   /* + tEnd.getTime() + " " */ + tEnd );

        // Now check if our task overlaps with previously scheduled tasks
        // (should NOT happen)
        final String msg = "Overlapping tasks, should never happen";
        for( Iterator iter = schedTasks.iterator(); iter.hasNext(); ) {
            
            TimelineTask sched = (TimelineTask) iter.next();
            Date sStart = sched.getStartTime();
            Date sEnd   = sched.getEndTime();

//            System.out.println( "Sched: "  + sched.getID()  );
//            System.out.println( "sStart: " /* + sStart.getTime() + " " */ + sStart );
//            System.out.println( "sEnd: "   /* + sEnd.getTime() + " " */ + sEnd );
            
            if( tStart.after( sStart) && tStart.before( sEnd )) {
                System.out.println( msg + ": start" );
                throw new RuntimeException( msg + ": start" );
            }
            if( tEnd.after( sStart ) && tEnd.before( sEnd )) {
                System.out.println( msg + ": end" );
                System.out.println( "Sched: " + sched.getID() );
                throw new RuntimeException( msg + ": end" );
            }
        }
        //===========================================================
        
        schedTasks.add( task );
    }
    
    
    /** @return    A summary view of the scheduled tasks */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append( "[" );
        TimelineTask[] tasks = getSchedTasks();
        for( int i = 0; i < tasks.length; i++ ) {
            if( i > 0 )
                sb.append( " " );
            TimelineTask task = tasks[i];
            sb.append( task.getID() )
              .append( "@" )
              .append( task.getStartTime().getTime() );
        }
        sb.append( "]" );
        return sb.toString();
    }
    
    /**
     * Overrides Object#equals().
     * 
     * Two Timelines are equal if the include the same scheduled
     * tasks with the same start times.
     * 
     * @param timeline The Timeline to compare to this one.
     */
    public boolean equals( Object timeline ) {
        
        Timeline other = (Timeline) timeline;
        TimelineTask[] ttasks = this.getSchedTasks();
        TimelineTask[] otasks = other.getSchedTasks();
        if( ttasks.length != otasks.length )
            return false;
        
        for( int i = 0; i < otasks.length; i++ ) {
            if( ttasks[i] != otasks[i] )
                return false;
            if( ttasks[i].getStartTime() != otasks[i].getStartTime() )
                return false;
        }
        return true;
    }
    
    /**
     * @return The duration of this timeline's range in milliseconds.
     */
    public long getRangeDuration() {
        PCFInterval interval = range.getIntervalAt( 1 );
        long ret = interval.getEnd().getTime() - interval.getStart().getTime();
        return ret;
    }

    // FOR TESTING ONLY
    void setTimeline( PCF timeline ) {
        this.timeline = timeline;
    }
}
