
package org.eso.ohs.phase2.visiplot;

import java.util.Date;
import java.util.Vector;

import javax.swing.JOptionPane;

import org.eso.ohs.core.gui.baseaction.IncrementalAction;
import org.eso.ohs.core.gui.widgets.ErrorMessages;
import org.eso.ohs.core.utilities.ObjectIOException;
import org.eso.ohs.dfs.BusinessObject;
import org.eso.ohs.dfs.ObservationBlock;
import org.eso.ohs.persistence.AppConfig;
import org.eso.ohs.persistence.Config;
import org.eso.ohs.persistence.Media;
import org.eso.ohs.persistence.ObjectManager;
import org.eso.ohs.persistence.ObjectNotFoundException;
import org.eso.ohs.phase2.orang.OBScheduleInfo;
import org.eso.ohs.phase2.orang.ObDataContainer;
import org.eso.ohs.phase2.visibility.VisibilityJCalc;
import org.eso.ohs.phase2.visibility.VisicalcOptions;
import org.eso.ohs.phase2.visibility.PCFPlotterFilter.ValueRejectedException;

/**
 * This class loads OBs, computes the visibilty and performs the necessary calculations on the 
 * Observation Blocks, in order to plot them with visiplots. 
 * 
 * @author pnunes 19.10.2005
 *
 */
public class IncrementalOBLoadingVisiCalc implements IncrementalAction {
	
    /**
	 * Member variables
	 */
	
	// object manager
	private ObjectManager objMgr = ObjectManager.getObjectManager();
	
		
	// counter upper limit
	private int upperLimit_ = 0;
		
	//	OB vector
	private Vector obVector_ = new Vector();
	
	// Loaded Observation blocks	
    private ObservationBlock[] loadedObs_ = null;
    
    // grouped OBs for plotting
    private ObDataContainer[][] groupedObs_ = null;        
    
    // Visiplot options
	private VisiPlotOptions visiplotOptions_;    
	
	// OBs with computed visibility
	private OBScheduleInfo [] computedOBs_;
	
	// observatory coordinates
	private double observLongit_;
	private double observLat_;
	
	// chart frame
	private ChartFrame frame_;
	
	// OB IDs to load
	private Long [] obIds_;
	
	// state of the class
	private boolean obsLoaded_;
	private boolean visibilitiesComputed_;
	private boolean chartBuilt_;
	private int numCalibBlocksFound_;

	
	/**
	 * Constructor
	 * 
	 * @param frame
	 * @param options
	 * @param obs
	 */	
	public IncrementalOBLoadingVisiCalc(VisiPlotOptions options, Long [] obIds, ChartFrame frame) {
		frame_ = frame;
		obIds_ = obIds;
		visiplotOptions_= options;
		upperLimit_ = obIds_.length;
        observLongit_ = AppConfig.getAppConfig().getLongitude();
    	observLat_ = AppConfig.getAppConfig().getLatitude();
	}
	

    public int getNumStepsRemaining(int counter) {
        return upperLimit_ - counter;
    }

    
    // calculate how many steps are to be performed
    // number of Obs loaded + number of addings to the chart frame
    public void startOfAction() {
    	obsLoaded_ = false;
    	visibilitiesComputed_ = false;
    	chartBuilt_ = false;    	    	
    	numCalibBlocksFound_ = 0;
    }

    public void endOfAction() { 
        frame_.displayPlots();
        frame_.pack();
        frame_.setVisible(true);
    }

    public void performStep(int counter) {
        if (counter < upperLimit_) {
            doStep(counter);
        }
    }
    
    
    
    /**
     * Performs the steps
     */
    void doStep(int ctr) {

    	// checks if the OBs are loaded already
    	if (! obsLoaded_) {    		
    		loadOBs(ctr);
    	} else { 
    		    		
    		if (! visibilitiesComputed_) {
    			
    			
    			// compute all the visibilities
    			computeVisibilities();
    			
    			// warn about calibration blocks
    			if (numCalibBlocksFound_ > 0) {
        			JOptionPane.showMessageDialog(null,
    						new Object[] {
    						"Visibility Plots does not plot Calibration Blocks.",
    						"These objects will be ignored.",
    				},
    				numCalibBlocksFound_ + " Calibration Block(s) selected",
    				JOptionPane.ERROR_MESSAGE
    				);
        		}
    			    			
    		} else
    			if (! chartBuilt_) {    				
    				//builds the chart step by step    				
    				buildChart(ctr); 
    				
    			}
    		
    	}
    }


	private void buildChart(int ctr) {
		// setting the index 
		int index = ctr - obIds_.length - 1;
		
		if (index < groupedObs_.length) {
			// build the chart
			String report[] = VisiplotUtils.getTextReport(groupedObs_[index], visiplotOptions_);    				    				
			
			// find the correct OBScheduleInfo object
			OBScheduleInfo obInfo = VisiplotUtils.find(groupedObs_[index][0].getObInfo().getId(), computedOBs_,visiplotOptions_);    				    		
			OBVisiplotInfo visi = new OBVisiplotInfo(obInfo, visiplotOptions_);
			
			frame_.addReport(report);
			frame_.addVisiplot(visi);
		}
	}


	private void computeVisibilities() {
		VisibilityJCalc visiCalc = new VisibilityJCalc();    			
		VisicalcOptions visicalcOptions = new VisicalcOptions(visiplotOptions_);
		int interval = visiplotOptions_.getIntValue(VisiPlotOptions.dataQueryIntKey);
		Date endTime = visiplotOptions_.getISODate(VisiPlotOptions.endTimeKey);
		Date startTime = visiplotOptions_.getISODate(VisiPlotOptions.startTimeKey);    
		
		ObDataContainer [] obContainers = ObDataContainer.createContainers(loadedObs_);
		
		try {
			computedOBs_ = visiCalc.calc(obContainers,interval,startTime,endTime,visicalcOptions,observLat_,observLongit_);
			visibilitiesComputed_ = true;
		} catch (ValueRejectedException e) {					
			e.printStackTrace();
		}
	}


	private void loadOBs(int ctr) {
		// load the OB step by step
		if (obIds_[ctr] != null) {
			ObservationBlock ob = getOB(obIds_[ctr]);
			if (ob != null) {
				obVector_.addElement(ob);
			} else {
				numCalibBlocksFound_ = numCalibBlocksFound_ + 1; 
			}
		}
		       		    		
		// check if it loaded all the OBs
		if ( ctr == obIds_.length - 1) {
			
			//  copy the vector to the array
			loadedObs_ = new ObservationBlock[obVector_.size()];
			obVector_.copyInto(loadedObs_);
			
			// groups OBs by target if applicable
			int numGroupedOBs = groupOBs();
			
			// sets the upperLimit_ variable:
			// number of OBs loaded
			// one step more to compute visibilities
			// plus number of grouped OB, to add to the plot
			
			upperLimit_ = upperLimit_ +  1 + numGroupedOBs;
			obsLoaded_ = true;
		}
	}
    
    /**
     * Groups the OBs by target depending on the visiplots options
     */
    private int groupOBs() {    	    	
    	    	
    	ObDataContainer [] containers = VisiplotUtils.getOBContainers(loadedObs_);
    	
    	groupedObs_ =
            VisiplotUtils.groupDuplicates(
            		containers,
                visiplotOptions_.isEnabled(VisiPlotOptions.groupPlotKey));
    	groupedObs_ = VisiplotUtils.sort(groupedObs_, visiplotOptions_);

        return (groupedObs_.length);
    }
    
    
    /**
	 * Loads an ObservationBlock from the database
	 * 
	 * @param id Database ID
	 * @return ObservationBlock business object
	 */
	private ObservationBlock getOB(Long id) {
		ObservationBlock rc = null;
		try {
			Media device = Config.getCfg().getMediaFromId(id.longValue());
			Class objType = Config.getCfg().getClassFromId(id.longValue()); 	       
			BusinessObject bo = (BusinessObject)objMgr.getBusObj(
					device, id.longValue(), objType);
			
			/**cmavros 5-11-01
			 * Add a warning message when a
			 * Calibration block is selected.
			 */
			if (bo instanceof ObservationBlock) {
				rc = (ObservationBlock)bo; 
			}
			
			else{				
					
				
			}
			
		} catch (ObjectNotFoundException e) {
			e.printStackTrace();
			ErrorMessages.announceNoObject(null, e);
			return null;
		} catch (ObjectIOException e) {
			ErrorMessages.announceIOError(null, e);
			return null;
		} //end try
		return rc;
	}
	
	
	
	
	
	
	

    
    

    
}